#!/usr/bin/env python3

#---------------------------------
# Author: Ankit Anand
# Date: 26/08/25
# Email: ankit0.anand0@gmail.com
#---------------------------------

from pathlib import Path
import matplotlib as mpl
import matplotlib.font_manager as fm
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.patches import Rectangle
import numpy as np
import fnmatch

#===== Loading Devanagari font ========
def _load_devanagari_font():
	"""
	Load devanagari font as it works for both English and Hindi.
	"""
	
	# Path to your bundled font
	font_path = Path(__file__).resolve().parents[1] / "fonts" / "NotoSansDevanagari-Regular.ttf"
	
	# Register the font with matplotlib
	fm.fontManager.addfont(str(font_path))
	
	# Get the font family name from the file
	hindi_font = fm.FontProperties(fname=str(font_path))
	
	# Set as default rcParam
	mpl.rcParams['font.family'] = hindi_font.get_name()

#==============

class Fig:
	"""
	A utility class that provides easy-to-use API for 
	plotting 1D/2D signals along with clean representations 
	of annotations, events.

	Parameters
	----------
	arrangement: str
		- Arrangement of the subplots you want in the figure.
		- "a" for aux plot which is good for adding annotations and events onto.
		- "s" for signal plot which is good for 1D array.
		- "m" for matrix plot which is good for 2D array.
		- E.g., "asm" or "aasm" or "saams", ...
		- Default: "asm"
	xlim: tuple[number, number] | None
		- Since all the subplots share x-axis, we set the x limit while creating the figure.
		- Default: None
	width: number
		- Width of the figure
		- Default: 16
	dark_mode: bool
		- Do you want dark mode?
		- Default: False
	abc: bool
		- Adds a, b, c, ... to the subplots for easier referencing.
		- Default: False
	fig_num: str | None
		- Adds a figure number to the top left of the figure.
		- Eg. "1.1"
		- Default: None
	"""
	
	def __init__(self, arrangement="asm", xlim=None, width=16, dark_mode=False, abc=False, fig_num=None):
		
		_load_devanagari_font()
		
		if dark_mode:
			plt.style.use("dark_background")
		else: # We do not reset to default as it changes other params leading to unexpected behaviour 
			color_keys = [
				"axes.facecolor", "axes.edgecolor", "axes.labelcolor",
				"xtick.color", "ytick.color", "text.color",
				"figure.facecolor", "figure.edgecolor",
				"legend.facecolor", "legend.edgecolor",
				"axes.prop_cycle", # This is to make sure that the plots colors are brighter in light mode
			]
			for k in color_keys:
				mpl.rcParams[k] = mpl.rcParamsDefault[k]
		
		self._xlim = xlim
		self._curr_row_idx = 1 # Starting from 1 because row 0 is reserved for reference subplot
		self._curr_color_idx = 0 # So that we have different color across all the subplots to avoid legend confusion
		self._abc = abc # This tells whether to add a, b, c, ... to the subplots for better referencing
		self._fig_num = fig_num # Add a figure number to the entire figure
		
		# Subplot setup
		self._fig, self._axs = self._generate_subplots(arrangement, width) # This will fill in the all the above variables
		
		
	def _get_curr_row(self):
		"""
		Get the active row where you can add
		either annotations or events.
		"""
		curr_row = self._axs[self._curr_row_idx]
		self._curr_row_idx += 1
		
		return curr_row
	
	def _get_prev_row(self):
		"""
		Get the prev row where you can add arrows.
		"""
		prev_pow = self._axs[self._curr_row_idx - 1]
		
		return prev_pow
	
	def _get_new_color(self):
		"""
		Get a new color for different lines.
		"""
		colors = plt.cm.tab20.colors
		self._curr_color_idx += 1
		
		return colors[self._curr_color_idx % len(colors)]
	
	def _calculate_extent(self, x, y, o):
		"""
		
		Parameters
		----------
		o: str
			- Origin
		"""
		# Handle spacing safely
		if len(x) > 1:
			dx = x[1] - x[0]
		else:
			dx = 1  # Default spacing for single value
		if len(y) > 1:
			dy = y[1] - y[0]
		else:
			dy = 1  # Default spacing for single value
			
		if o == "lower":
			return  [x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2]
		else:
			return [x[0] - dx / 2, x[-1] + dx / 2, y[-1] + dy / 2, y[0] - dy / 2]
	
	
	def _generate_subplots(self, arrangement, width):
		"""
		Generate subplots based on the configuration.
		"""
		
		xlim = self._xlim
		fig_width = width
		
		n_aux_sp = arrangement.count("a")
		n_signal_sp = arrangement.count("s")
		n_matrix_sp = arrangement.count("m")
		n_sp = 1 + n_aux_sp + n_signal_sp + n_matrix_sp # +1 is for the first reference subplot
		
		# Decide heights of different subplots type
		height = {}
		height["r"] = 0.0 # Reference height
		height["a"] = 0.4 # Aux height
		height["s"] = 2.0 # Signal height
		height["m"] = 4.0 # Matrix height
		cbar_width = 0.01
		
		arrangement = "r" + arrangement # "r" is to include the reference
		
		# Calculate height ratios list based on the arrangement
		for char in arrangement:
			height_ratios = [height[char] for char in arrangement]
			
		# Calculate total fig height
		fig_height = height["r"] + (n_aux_sp * height["a"]) + (n_signal_sp * height["s"]) + (n_matrix_sp * height["m"])
		
		# Create figure and axs
		fig, axs = plt.subplots(n_sp, 2, figsize=(fig_width, fig_height), height_ratios=height_ratios, width_ratios=[1, cbar_width])
		
		for i, char in enumerate(arrangement): # For each of the subplots, we modify the layout accordingly
			if char == "r":
				axs[i, 0].axis("off")
				axs[i, 1].axis("off")
			elif char == "a": # Remove ticks and labels from all the aux subplots
				axs[i, 0].tick_params(left=False, bottom=False, labelleft=False, labelbottom=False)
				axs[i, 1].axis("off")
			elif char == "s":
				axs[i, 0].tick_params(bottom=False, labelbottom=False)
				axs[i, 1].axis("off")
			elif char == "m":
				axs[i, 0].tick_params(bottom=False, labelbottom=False)
				
			axs[i, 0].sharex(axs[0, 0])
		
		# Add subplot labels (a, b, c, d, ...) for better referencing.
		if self._abc is True:
			label_counter = 0
			for i, char in enumerate(arrangement):
				if char != "r":  # Skip the reference subplot
					label = chr(97 + label_counter)  # 97 is ASCII for 'a'
					axs[i, 0].text(-0.06, 0.5, f'({label})', transform=axs[i, 0].transAxes, fontsize=12, fontweight='bold', va='center', ha='right')
					label_counter += 1
		
		# Add figure number at top-left
		if self._fig_num is not None:
			fig.text(0.12, 0.9, f'fig - {self._fig_num}', fontsize=10, fontweight='bold', va='top', ha='left', color='black', bbox=dict(boxstyle='round,pad=0.3', facecolor='lightgray', edgecolor='gray', linewidth=1.0))
		
		axs[-1, 0].tick_params(bottom=True, labelbottom=True)
		
		# Add the figure title on top-left (if any)
		if self._fig_num is not None:
			fig.suptitle(f'fig - {self._fig_num}', fontsize=12, fontweight='bold', x=0.01, ha='left', va='top', y=0.98)
			
		# xlim should be applied on reference subplot, rest all subplots will automatically adjust
		if xlim is not None:
			axs[0, 0].set_xlim(xlim)
			
		fig.subplots_adjust(hspace=0.2, wspace=0.05)
		
		return fig, axs
	
	def add_signal(self, y, x=None, c=None, ls=None, lw=None, m=None, ms=3, label=None, ylabel=None, ylim=None, yticks=None, yticklabels=None, xticks=None, xticklabels=None, grid=True, ax=None):
		"""
		Add signal to the figure.
			
		Parameters
		----------
		y: np.ndarray
			- Signal y values.
		x: np.ndarray | None
			- Signal x values.
			- Default: None (indices will be used)
		c: str
			- Color of the line.
			- Default: None
		ls: str
			- Linestyle
			- Default: None
		lw: Number
			- Linewidth
			- Default: None
		m: str
			- Marker
			- Default: None
		ms: number
			- Markersize
			- Default: 3
		label: str
			- Label for the plot.
			- Legend will use this.
			- Default: None
		ylabel: str
			- y-label for the plot.
			- Default: None
		ylim: tuple
			- y-lim for the plot.
			- Default: None
		yticks: Arraylike
			- Positions at which to place y-axis ticks.
		yticklabels : list of str, optional
			- Labels corresponding to `yticks`. Must be the same length as `yticks`.
		xticks: Arraylike
			- Positions at which to place x-axis ticks.
		xticklabels : list of str, optional
			- Labels corresponding to `xticks`. Must be the same length as `xticks`.
		grid: bool
			- Do you want the grid?
			- Default: True
		ax: int
			- Which specific axis to plot (1, 2, 3, ...)
			- None

		Returns
		-------
		None
		"""
		
		curr_row = self._get_curr_row() if ax is None else self._axs[ax]
		
		if x is None: 
			x = np.arange(y.size)
		
		if c is None: 
			c = self._get_new_color()
		
		curr_row[0].plot(x, y, color=c, linestyle=ls, linewidth=lw, marker=m, markersize=ms, label=label)
		
		if ylabel is not None: 
			curr_row[0].set_ylabel(ylabel)
		
		if ylim is not None: 
			curr_row[0].set_ylim(ylim)
		
		if yticks is not None:
			curr_row[0].set_yticks(yticks)
			if yticklabels is not None:
				curr_row[0].set_yticklabels(yticklabels)
				
		if xticks is not None:
			curr_row[0].set_xticks(xticks)
			if xticklabels is not None:
				curr_row[0].set_xticklabels(xticklabels)
		
		if grid is True:
			curr_row[0].grid(True, linestyle='--', linewidth=0.7, color="lightgray" ,alpha=0.6)
				
		
	def add_matrix(self, M, y=None, x=None, c="viridis", o="upper", label=None, ylabel=None, ylim=None, yticks=None, yticklabels=None, xticks=None, xticklabels=None, cbar=True, grid=True, alpha=1, ax=None):
		"""
		Add matrix to the figure.
			
		Parameters
		----------
		M: np.ndarray
			- Matrix (2D) array
		y: np.ndarray | None
			- y axis values.
		x: np.ndarray | None (indices will be used)
			- x axis values.
			- Default: None (indices will be used)
		c: str
			- cmap for the matrix.
			- Default: None
		o: str
			- origin
			- Default: "lower"
		label: str
			- Label for the plot.
			- Legend will use this.
			- Default: None
		ylabel: str
			- y-label for the plot.
			- Default: None
		ylim: tuple
			- y-lim for the plot.
			- Default: None
		yticks: Arraylike
			- Positions at which to place y-axis ticks.
		yticklabels : list of str, optional
			- Labels corresponding to `yticks`. Must be the same length as `yticks`.
		xticks: Arraylike
			- Positions at which to place x-axis ticks.
		xticklabels : list of str, optional
			- Labels corresponding to `xticks`. Must be the same length as `xticks`.
		cbar: bool
			- Show colorbar
			- Default: True
		grid: bool
			- Do you want the grid?
			- Default: True
		alpha: float (0 to 1)
			- Transparency level
			- 1 being opaque and 0 being completely transparent
			- Default: 1
		ax: int
			- Which specific axis to plot (1, 2, 3, ...)
			- None
		
		Returns
		-------
		None
		"""
		if x is None: x = np.arange(M.shape[1])
		if y is None: y = np.arange(M.shape[0])
		
		curr_row = self._get_curr_row() if ax is None else self._axs[ax]
		
		extent = self._calculate_extent(x, y, o)
		
		im = curr_row[0].imshow(M, aspect="auto", origin=o, cmap=c, extent=extent, alpha=alpha)
		
		if ylabel is not None: curr_row[0].set_ylabel(ylabel)
		
		if ylim is not None:
			if o == "lower":
				curr_row[0].set_ylim(ylim)
			elif o == "upper":
				curr_row[0].set_ylim(ylim[::-1])
		
		if cbar is True:
			cbar = plt.colorbar(im, cax=curr_row[1])
			if label is not None:
				cbar.set_label(label, labelpad=5)
		
		if yticks is not None:
			curr_row[0].set_yticks(yticks)
			if yticklabels is not None:
				curr_row[0].set_yticklabels(yticklabels)
				
		if xticks is not None:
			curr_row[0].set_xticks(xticks)
			if xticklabels is not None:
				curr_row[0].set_xticklabels(xticklabels)
				
		if grid is True:
			curr_row[0].grid(True, linestyle='--', linewidth=0.7, color="lightgray" ,alpha=0.6)
			
				
				
	def add_events(self, events, c=None, ls=None, lw=None, label=None, grid=True, ax=None):
		"""
		Add events to the figure.
		
		Parameters
		----------
		events: np.ndarray
			- All the event marker values.
		c: str
			- Color of the event marker.
			- Default: "k"
		ls: str
			- Line style.
			- Default: "-"
		lw: float
			- Linewidth.
			- Default: 1.5
		label: str
			- Label for the event type.
			- This will appear in the legend.
			- Default: None
		grid: bool
			- Do you want the grid?
			- Default: True
		ax: int
			- Which specific axis to plot (1, 2, 3, ...)
			- None

		Returns
		-------
		None
		"""
		
		curr_row = self._get_curr_row() if ax is None else self._axs[ax]
		
		if c is None: c = self._get_new_color()
		
		xlim = self._xlim
		
		for i, event in enumerate(events):
			if xlim is not None:
				if xlim[0] <= event <= xlim[1]:
					if i == 0: # Label should be set only once for all the events
						curr_row[0].axvline(x=event, color=c, linestyle=ls, linewidth=lw, label=label)
					else:
						curr_row[0].axvline(x=event, color=c, linestyle=ls, linewidth=lw)
			else:
				if i == 0: # Label should be set only once for all the events
					curr_row[0].axvline(x=event, color=c, linestyle=ls, linewidth=lw, label=label)
				else:
					curr_row[0].axvline(x=event, color=c, linestyle=ls, linewidth=lw)
					
		if grid is True:
			curr_row[0].grid(True, linestyle='--', linewidth=0.7, color="lightgray" ,alpha=0.6)
					
	def add_annotation(self, ann, label=None, patterns=None, ylim=(0, 1), text_loc="m", grid=True, ax=None):
		"""
		Add annotation to the figure.
		
		Parameters
		----------
		ann : list[tuple[Number, Number, str]] | None
			- A list of annotation spans. Each tuple should be (start, end, label).
			- Default: None (no annotations).
		label: str
			- Label for the annotation type.
			- This will appear to the right of the aux plot.
			- Default: None
		patterns: list[str]
			- Patterns to group annotations
			- E.g., "*R" or "<tag>*" or ["A*", "*B"]
			- All elements in a group will have same color.
		ylim: tuple[number, number]
			- Y-limit for the annotation.
			- Default: (0, 1)
		text_loc: str
			- Location of text relative to the box. (b for bottom, m for middle, t for top)
			- Default: "m"
		grid: bool
			- Do you want the grid?
			- Default: True
		ax: int
			- Which specific axis to plot (1, 2, 3, ...)
			- None
		Returns
		-------
		None
		"""
		curr_row = self._get_curr_row() if ax is None else self._axs[ax]
		
		xlim = self._xlim
		
		if isinstance(patterns, str): patterns = [patterns]
		ann_copy = ann.copy()
		
		if patterns is not None:
			for i, (start, end, tag) in enumerate(ann_copy):
				group = None
				for j, pattern in enumerate(patterns):
					if fnmatch.fnmatch(tag, pattern):
						group = j
						break
				ann_copy[i] = (start, end, tag, group)
		else:
			for i, (start, end, tag) in enumerate(ann_copy):
				ann_copy[i] = (start, end, tag, None)
					
		colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
		
		# Text Location
		if text_loc in ["b", "bottom", "lower", "l"]: 
			text_yloc = ylim[0] + 0.1 * (ylim[1] - ylim[0])
		elif text_loc in ["t", "top", "u", "upper"]:
			text_yloc = ylim[1] - 0.1 * (ylim[1] - ylim[0])
		else:
			text_yloc = (ylim[1] + ylim[0]) / 2
		
		for i, (start, end, tag, group) in enumerate(ann_copy):
			# We make sure that we only plot annotation that are within the x range of the current view
			if xlim is not None:
				if start >= xlim[1] or end <= xlim[0]:
					continue
				
				# Clip boundaries to xlim
				start = max(start, xlim[0])
				end = min(end, xlim[1])
			
				
				if group is not None:
					box_color = colors[group]
				else:
					box_color = "lightgray"
				
				width = end - start
				rect = Rectangle((start, ylim[0]), width, ylim[1] - ylim[0], facecolor=box_color, edgecolor="black", alpha=0.7)
				curr_row[0].add_patch(rect)
				
				text_obj = curr_row[0].text(
					(start + end) / 2, text_yloc, tag,
					ha='center', va='center',
					fontsize=10, color="black", fontweight='bold', zorder=10, clip_on=True
				)
				
				text_obj.set_clip_path(rect)
			else:
				if group is not None:
					box_color = colors[group]
				else:
					box_color = "lightgray"
				
				width = end - start
				rect = Rectangle((start, ylim[0]), width, ylim[1] - ylim[0], facecolor=box_color, edgecolor="black", alpha=0.7)
				curr_row[0].add_patch(rect)
				
				text_obj = curr_row[0].text(
					(start + end) / 2, text_yloc, tag,
					ha='center', va='center',
					fontsize=10, color="black", fontweight='bold', zorder=10, clip_on=True
				)
				
				text_obj.set_clip_path(rect)
				
		if label is not None:
			curr_row[0].set_ylabel(label, rotation=0, ha="center", va="center")
			curr_row[0].yaxis.set_label_position("right")
			curr_row[0].yaxis.set_label_coords(1.05, 0.75)
		
		if grid is True:
			curr_row[0].grid(True, linestyle='--', linewidth=0.7, color="lightgray" ,alpha=0.6)
	
	def add_arrows(self, xys, labels, text_offset=(0, 0), c="r", fontsize=12, ax=None):
		"""
		Add multiple arrows pointing to specific points with boxed labels at the tails.
	
		Parameters
		----------
		xys : list[tuple[float, float]] | tuple[float, float]
			- List of target points (x, y) for the arrow heads.
		labels : list[str] | str
			- List of text labels at the arrow tails.
			- If str, the same label is used for all points.
		text_offset : tuple[float, float] | list[tuple[float, float]]
			- Offset(s) (dx, dy) for label positions from arrow tails.
			- If single tuple, same offset is applied to all.
		c : str | list[str]
			- Color(s) for arrow and text.
			- If str, same color is applied to all.
		fontsize : int | list[int]
			- Font size(s) of the label text.
			- If int, same size is applied to all.
		ax : int | None
			- Which specific axis to plot (1, 2, 3, ...).
			- If None, uses the current row.
	
		Returns
		-------
		None
		"""
		curr_row = self._get_prev_row() if ax is None else self._axs[ax]
		
		# Normalize single values into lists
		if isinstance(xys, tuple):
			xys = [xys]
		n = len(xys)
		if isinstance(labels, str):
			labels = [labels] * n
		if isinstance(text_offset, tuple):
			text_offset = [text_offset] * n
		if isinstance(c, str):
			c = [c] * n
		if isinstance(fontsize, int):
			fontsize = [fontsize] * n
			
		for (xy, label, offset, color, fs) in zip(xys, labels, text_offset, c, fontsize):
			arrowprops = dict(arrowstyle="->", color=color, lw=2)
			bbox = dict(boxstyle="round,pad=0.3", fc="white", ec=color, lw=1.2)
			
			text_x, text_y = xy[0] + offset[0], xy[1] + offset[1]
			
			curr_row[0].annotate(
				label,
				xy=xy, xycoords="data",
				xytext=(text_x, text_y), textcoords="data",
				arrowprops=arrowprops,
				fontsize=fs,
				color=color,
				ha="center", va="center",
				bbox=bbox
			)
			
	def add_legend(self, ypos=1.0):
		"""
		Add legend to the figure.

		Parameters
		----------
		ypos: float
			- y position from the top.
			- > 1 to push it higher, < 1 to push it lower
			- Default: 1.3
		
		Returns
		-------
		None
		"""
		axs = self._axs
		fig = self._fig
		
		all_handles, all_labels = [], []
		
		for ax in axs:
			handles, labels = ax[0].get_legend_handles_labels()
			all_handles.extend(handles)
			all_labels.extend(labels)
			
		# remove duplicates if needed
		fig.legend(all_handles, all_labels, loc='upper right', bbox_to_anchor=(0.95, ypos), ncol=3, frameon=True, bbox_transform=fig.transFigure)
		
	def add_title(self, title=None, s=13):
		"""
		Add title to the figure.

		Parameters
		----------
		title: str | None
			- Title of the figure.
			- Default: None
		s: Number
			- Font size.
			- Default: None
		"""
		axs = self._axs
		ref_ax = axs[0, 0] # Title is added to the top subplot (ref subplot)
		
		if title is not None:
			ref_ax.set_title(title, pad=10, size=s)
			
			
	def add_xlabel(self, xlabel=None, s=None):
		"""
		Add shared x-label to the figure.

		Parameters
		----------
		xlabel: str | None
			- xlabel for the figure.
			- Default: None
		s: Number
			- Font size.
			- Default: None
		"""
		axs = self._axs
		ref_ax = axs[-1, 0] # X-label is added to the last subplot
		if xlabel is not None:
			ref_ax.set_xlabel(xlabel, size=s)
			
	def add_xticks(self, xticks=None):
		"""
		Not implemented yet
		"""
		raise NotImplementedError("Please raise a github issue https://github.com/meluron-toolbox/modusa/issues")
		
	def save(self, path="./figure.png"):
		"""
		Save the figure.

		Parameters
		----------
		path: str
			- Path to the output file.

		Returns
		-------
		None
		"""
		fig = self._fig
		fig.savefig(path, bbox_inches="tight")
		
#======== Distribution ==========
def hill_plot(*args, labels=None, xlabel=None, ylabel=None, title=None, widths=0.7, bw_method=0.3, jitter_amount=0.1, side='upper', show_stats=True, ax=None):
		"""
		A plot to see distribution of different groups
		along with statistical markers.
		
		Parameters
		----------
		*args: array-like
				- Data arrays for each group.
		labels: list of str, optional
				- Labels for each group (y-axis).
		ylabel: str, optional
				Label for y-axis.
		xlabel: str, optional
				Label for x-axis.
		title: str, optional
				Plot title.
		widths: float, optional
				Width of violins.
		bw_method: float, optional
				Bandwidth method for KDE.
		jitter_amount: float, optional
				Amount of vertical jitter for strip points.
		side: str, 'upper' or 'lower'
				Which half of the violin to draw (upper or lower relative to y-axis).
		show_stats: bool, optional
				Whether to show mean and median markers.
		ax: matplotlib axes, optional
				Axes to plot on.
		"""
	
		plt.style.use("default") # Not supporting dark mode
		plt.rcParams['font.family'] = "DejaVu Sans" # Devnagari not needed for this.
		
		created_fig = False
		if ax is None:
			fig, ax = plt.subplots(figsize=(8, len(args) * 1.5))
			created_fig = True
			
		n = len(args)
	
		# Default labels/colors
		if labels is None:
			labels = [f"Group {i}" for i in range(1, n+1)]
		if isinstance(labels, str):
			labels = [labels]
			
		colors = plt.cm.tab10.colors
		if len(colors) < n:
			colors = [colors[i % len(colors)] for i in range(n)] # Repeat colors
			
		# --- Half-violin ---
		parts = ax.violinplot(args, vert=False, showmeans=False, showmedians=False, widths=widths, bw_method=bw_method)
	
		# Remove the default bar lines from violin plot
		for key in ["cbars", "cmins", "cmaxes", "cmedians"]:
			if key in parts:
				parts[key].set_visible(False)
					
		# Clip violin bodies to show only upper or lower half
		for i, pc in enumerate(parts['bodies']):
			verts = pc.get_paths()[0].vertices
			y_center = i + 1  # Center y-position for this violin
		
			if side == "upper":
					verts[:, 1] = np.maximum(verts[:, 1], y_center)
			else:  # 'lower'
					verts[:, 1] = np.minimum(verts[:, 1], y_center)
		
			pc.set_facecolor(colors[i])
			pc.set_edgecolor(colors[i])
			pc.set_linewidth(1.5)
			pc.set_alpha(0.3)
			
		# --- Strip points with jitter ---
		for i, x in enumerate(args, start=1):
			x = np.array(x)
			jitter = (np.random.rand(len(x)) - 0.5) * jitter_amount
			y_positions = np.full(len(x), i) + jitter
			ax.scatter(x, y_positions, color=colors[i-1], alpha=0.6, s=25, edgecolor="white", linewidth=0.8, zorder=2)
			
		# --- Statistical markers on violin distribution curve ---
		if show_stats:
			for i, (pc, x) in enumerate(zip(parts['bodies'], args), start=1):
				x = np.array(x)
				median_val = np.median(x)
				mean_val = np.mean(x)
				std_val = np.std(x)
			
				# Get the violin curve vertices
				verts = pc.get_paths()[0].vertices
			
				# Find y-position on violin curve for median
				median_mask = np.abs(verts[:, 0] - median_val) < (np.ptp(x) * 0.01)
				if median_mask.any():
					median_y = np.max(verts[median_mask, 1]) if side == "upper" else np.min(verts[median_mask, 1])
				else:
					median_y = i + widths/2 if side == "upper" else i - widths/2
					
				# Find y-position on violin curve for mean
				mean_mask = np.abs(verts[:, 0] - mean_val) < (np.ptp(x) * 0.01)
				if mean_mask.any():
					mean_y = np.max(verts[mean_mask, 1]) if side == "upper" else np.min(verts[mean_mask, 1])
				else:
					mean_y = i + widths/2 if side == "upper" else i - widths/2
					
				# Triangle offset from curve
				triangle_offset = 0.05
			
				# Mean marker - triangle below curve pointing up
				ax.scatter(mean_val, mean_y - triangle_offset, marker="^", s=30, 
									facecolor=colors[i-1], edgecolor="black", 
									linewidth=0.5, zorder=6,
									label="Mean" if i == 1 else "")
			
				# Mean value text - below the triangle
				ax.text(mean_val, mean_y - triangle_offset - 0.07, f"mean: {mean_val:.2f} ± {std_val:.2f}", ha="center", va="top", fontsize=8, color="black", zorder=7)
			
				# Median marker - triangle above curve pointing down
				ax.scatter(median_val, median_y + triangle_offset, marker="v", s=30, 
									facecolor=colors[i-1], edgecolor="black", 
									linewidth=0.5, zorder=6,
									label="Median" if i == 1 else "")
			
				# Median value text - above the triangle
				ax.text(median_val, median_y + triangle_offset + 0.07, f"median: {median_val:.2f}", ha="center", va="bottom", fontsize=8, color="black", zorder=7)
					
		# --- Labels & formatting ---
		ax.set_yticks(range(1, n + 1))
		ax.set_yticklabels(labels, fontsize=9)
		ax.tick_params(axis='x', labelsize=9)
		
		if side == "lower":
			ax.set_ylim(0.2, n + 0.5)
		else:
			ax.set_ylim(0.5, n + 0.5) 
			
		# Style improvements
		ax.spines['top'].set_visible(False)
		ax.spines['right'].set_visible(False)
		ax.grid(axis='x', alpha=0.3, linestyle="--", linewidth=0.5)
	
		if xlabel:
			ax.set_xlabel(xlabel, fontsize=9)
		if ylabel:
			ax.set_ylabel(ylabel, fontsize=9)
		if title:
			ax.set_title(title, fontsize=10, pad=20)
			
		plt.tight_layout()
		plt.close()
		return fig


def dist_plot(*args, ann=None, xlim=None, ylim=None, ylabel=None, xlabel=None, title=None, legend=None, show_hist=True, npoints=200, bins=30):
		"""
		Plot distribution.

		.. code-block:: python
			
			import modusa as ms
			import numpy as np
			np.random.seed(42)
			data = np.random.normal(loc=1, scale=1, size=1000)
			ms.plot_dist(data, data+5, data-10, ann=[(0, 1, "A")], legend=("D1", "D2", "D3"), ylim=(0, 1), xlabel="X", ylabel="Counts", title="Distribution")

		Parameters
		----------
		*args: ndarray
			- Data arrays for which distribution needs to be plotted.
			- Arrays will be flattened.
		ann : list[tuple[Number, Number, str] | None
			- A list of annotations to mark specific points. Each tuple should be of the form (start, end, label).
			- Default: None => No annotation.
		events : list[Number] | None
			- A list of x-values where vertical lines (event markers) will be drawn.
			- Default: None
		xlim : tuple[Number, Number] | None
			- Limits for the x-axis as (xmin, xmax).
			- Default: None
		ylim : tuple[Number, Number] | None
			- Limits for the y-axis as (ymin, ymax).
			- Default: None
		xlabel : str | None
			- Label for the x-axis.
			- - Default: None
		ylabel : str | None
			- Label for the y-axis.
			- Default: None
		title : str | None
			- Title of the plot.
			- Default: None
		legend : list[str] | None
			- List of legend labels corresponding to each signal if plotting multiple distributions.
			- Default: None
		show_hist: bool
			- Want to show histogram as well.
		npoints: int
			- Number of points for which gaussian needs to be computed between min and max.
			- Higher value means more points are evaluated with the fitted gaussian, thereby higher resolution.
		bins: int
			- The number of bins for histogram.
			- This is used only to plot the histogram.

		Returns
		-------
		plt.Figure
			- Matplotlib figure.
		"""
		from scipy.stats import gaussian_kde
	
		if isinstance(legend, str):
			legend = (legend, )
			
		if legend is not None:
			if len(legend) < len(args):
				raise ValueError(f"Legend should be provided for each signal.")
				
		# Create figure
		fig = plt.figure(figsize=(16, 4))
		gs = gridspec.GridSpec(2, 1, height_ratios=[0.1, 1])
	
		colors = plt.get_cmap('tab10').colors
	
		dist_ax = fig.add_subplot(gs[1, 0])
		annotation_ax = fig.add_subplot(gs[0, 0], sharex=dist_ax)
	
		# Set limits
		if xlim is not None:
			dist_ax.set_xlim(xlim)
			
		if ylim is not None:
			dist_ax.set_ylim(ylim)
			
		# Add plot
		for i, data in enumerate(args):
			# Fit gaussian to the data
			kde = gaussian_kde(data)
			
			# Create points to evaluate KDE
			x = np.linspace(np.min(data), np.max(data), npoints)
			y = kde(x)
			
			if legend is not None:
				dist_ax.plot(x, y, color=colors[i], label=legend[i])
				if show_hist is True:
					dist_ax.hist(data, bins=bins, density=True, alpha=0.3, facecolor=colors[i], edgecolor='black', label=legend[i])
			else:
				dist_ax.plot(x, y, color=colors[i])
				if show_hist is True:
					dist_ax.hist(data, bins=bins, density=True, alpha=0.3, facecolor=colors[i], edgecolor='black')
					
		# Add annotations
		if ann is not None:
			annotation_ax.set_ylim(0, 1) # For consistent layout
			# Determine visible x-range
			x_view_min = xlim[0] if xlim is not None else np.min(x)
			x_view_max = xlim[1] if xlim is not None else np.max(x)
			for i, (start, end, tag) in enumerate(ann):
				# We make sure that we only plot annotation that are within the x range of the current view
				if start >= x_view_max or end <= x_view_min:
					continue
				
				# Clip boundaries to xlim
				start = max(start, x_view_min)
				end = min(end, x_view_max)
				
				color = colors[i % len(colors)]
				width = end - start
				rect = Rectangle((start, 0), width, 1, color=color, alpha=0.7)
				annotation_ax.add_patch(rect)
				
				text_obj = annotation_ax.text((start + end) / 2, 0.5, tag, ha='center', va='center', fontsize=10, color='white', fontweight='bold', zorder=10, clip_on=True)
				text_obj.set_clip_path(rect)
				
		# Add legend
		if legend is not None:
			handles, labels = dist_ax.get_legend_handles_labels()
			fig.legend(handles, labels, loc='upper right', bbox_to_anchor=(0.9, 1.1), ncol=len(legend), frameon=True)
			
		# Set title, labels
		if title is not None:
			annotation_ax.set_title(title, pad=10, size=11)
		if xlabel is not None:
			dist_ax.set_xlabel(xlabel)
		if ylabel is not None:
			dist_ax.set_ylabel(ylabel)
			
		# Remove the boundaries and ticks from annotation axis
		if ann is not None:
			annotation_ax.tick_params(left=False, bottom=False, labelleft=False, labelbottom=False)
		else:
			annotation_ax.axis("off")
			
		fig.subplots_adjust(hspace=0.01, wspace=0.05)
		plt.close()
		return fig
	

# ======== Single Plot ===========
def plot(*args):
	"""
	To create 1D/2D plot for immediate
	inspection.

	Parameters
	----------
	*args: ndarray
		- Arrays to be plotted.
	"""
	n = len(args)
	
	if n < 1:
		raise ValueError("Need atleast 1 positional argument")
	
	## 1D Array
	if args[0].ndim == 1:
		if n > 1: # User also passed the xs
			fig = Fig("s").add_signal(args[0], args[1])
		else:
			fig = Fig("s").add_signal(args[0])
	
	## 2D Array
	if args[0].ndim == 2:
		if n == 1:
			fig = Fig("m").add_matrix(args[0])
		elif n == 2: # User also passed the xs
			fig = Fig("m").add_matrix(args[0], args[1])
		elif n == 3:
			fig = Fig("m").add_matrix(args[0], args[1], args[2])
	
	return fig
	
	
		
	
	
		
	


#!/usr/bin/env python3

from modusa import excp
from modusa.tools.base import ModusaTool
from typing import Any
import numpy as np

class MathOps(ModusaTool):
	"""
	Performs arithmetic and NumPy-style ops.

	Note
	----
	- Shape-changing operations like reshape, transpose, etc. are not yet supported. Use only element-wise or aggregation ops for now. 
	- Index alignment must be handled carefully in future extensions.
	"""
	
	def _axes_match(a1: tuple[np.ndarray, ...], a2: tuple[np.ndarray, ...]) -> bool:
		"""
		To check if two axes are same.

		It checks the length of the axes and the corresponding values.
		"""
		if len(a1) != len(a2):
			return False
		return all(np.allclose(x, y, atol=1e-8) for x, y in zip(a1, a2))
	
	
	#----------------------------------
	# To handle basic element wise 
	# math operations like
	# +, -, *, **, / ...
	#----------------------------------
	
	@staticmethod
	def add(a: Any, b: Any) -> np.generic | np.ndarray:
		try:
			result = np.add(a, b)
		except Exception as e:
			raise excp.InputError(f"`a` and `b` can't be added") from e
			
		if isinstance(a, str) and isinstance(b, str): # numpy actually concatenates, we do not want that
			raise excp.InputError(f"`a` and `b` can't be added")
		return result
	
	@staticmethod
	def subtract(a: Any, b: Any) -> np.generic | np.ndarray:
		try:
			result = np.subtract(a, b)
		except Exception as e:
			raise excp.InputError(f"`a` and `b` can't be subtracted") from e
		return result
		
	@staticmethod
	def multiply(a: Any, b: Any) -> np.generic | np.ndarray:
		try:
			result = np.multiply(a, b)
		except Exception as e:
			raise excp.InputError(f"`a` and `b` can't be multiplied") from e
		
		if isinstance(a, str) and isinstance(b, str): # numpy actually concatenates, we do not want that
			raise excp.InputError(f"`a` and `b` can't be multiplied")
		return result

	@staticmethod
	def divide(a: Any, b: Any) -> np.generic | np.ndarray:
		try:
			result = np.divide(a, b)
		except Exception as e:
			raise excp.InputError(f"`a` and `b` can't be divided") from e
		return result

	@staticmethod
	def power(a: Any, b: Any) -> np.generic | np.ndarray:
		try:
			result = np.power(a, b)
		except Exception as e:
			raise excp.InputError(f"`a` can't be exponentiated with `b`") from e
		return result
	
	@staticmethod
	def floor_divide(a: Any, b: Any) -> np.generic | np.ndarray:
		try:
			result = np.floor_divide(a, b)
		except Exception as e:
			raise excp.InputError(f"`a` can't be floor divided by `b`") from e
		return result
		
	#----------------------------------
	# To handle numpy aggregator ops
	#----------------------------------
	@staticmethod
	def mean(a: Any, axis: int | None = None) -> np.generic | np.ndarray:
		try:
			result = np.mean(a, axis=axis)
		except Exception as e:
			raise excp.InputError(f"can't find mean for `a`") from e
		return result
	
	@staticmethod
	def std(a: Any, axis: int | None = None) -> np.generic | np.ndarray:
		""""""
		try:
			result = np.std(a, axis=axis)
		except Exception as e:
			raise excp.InputError(f"can't find std for `a`") from e
		return result
	
	@staticmethod
	def min(a: Any, axis: int | None = None) -> np.generic | np.ndarray:
		try:
			result = np.min(a, axis=axis)
		except Exception as e:
			raise excp.InputError(f"can't find min for `a`") from e
		return result
	
	@staticmethod
	def max(a: Any, axis: int | None = None) -> np.generic | np.ndarray:
		try:
			result = np.max(a, axis=axis)
		except Exception as e:
			raise excp.InputError(f"can't find max for `a`") from e
		return result
	
	@staticmethod
	def sum(a: Any, axis: int | None = None) -> np.generic | np.ndarray:
		try:
			result = np.sum(a, axis=axis)
		except Exception as e:
			raise excp.InputError(f"can't find sum for `a`") from e
		return result
	
	#----------------------------------
	# To handle numpy ops where the
	# shapes are unaltered
	# sin, cos, exp, log, ...
	#----------------------------------
	
	@staticmethod
	def sin(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.sin(a)
		except Exception as e:
			raise excp.InputError(f"can't find sin for `a`") from e
		return result
	
	@staticmethod
	def cos(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.cos(a)
		except Exception as e:
			raise excp.InputError(f"can't find cos for `a`") from e
		return result
	
	@staticmethod
	def tanh(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.tanh(a)
		except Exception as e:
			raise excp.InputError(f"can't find tanh for `a`") from e
		return result
	
	@staticmethod
	def exp(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.exp(a)
		except Exception as e:
			raise excp.InputError(f"can't find exp for `a`") from e
		return result
	
	@staticmethod
	def log(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.log(a)
		except Exception as e:
			raise excp.InputError(f"can't find log for `a`") from e
		return result
	
	@staticmethod
	def log10(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.log10(a)
		except Exception as e:
			raise excp.InputError(f"can't find log10 for `a`") from e
		return result
	
	@staticmethod
	def log2(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.log2(a)
		except Exception as e:
			raise excp.InputError(f"can't find log2 for `a`") from e
		return result
	
	@staticmethod
	def log1p(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.log1p(a)
		except Exception as e:
			raise excp.InputError(f"can't find log1p for `a`") from e
		return result
	
	
	@staticmethod
	def sqrt(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.sqrt(a)
		except Exception as e:
			raise excp.InputError(f"can't find sqrt for `a`") from e
		return result
	
	@staticmethod
	def abs(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.abs(a)
		except Exception as e:
			raise excp.InputError(f"can't find abs for `a`") from e
		return result
	
	@staticmethod
	def floor(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.floor(a)
		except Exception as e:
			raise excp.InputError(f"can't find floor for `a`") from e
		return result
	
	@staticmethod
	def ceil(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.ceil(a)
		except Exception as e:
			raise excp.InputError(f"can't find ceil for `a`") from e
		return result
	
	@staticmethod
	def round(a: Any) -> np.generic | np.ndarray:
		try:
			result = np.round(a)
		except Exception as e:
			raise excp.InputError(f"can't find round for `a`") from e
		return result
	
	#------------------------------------
	# TODO: Add shape-changing ops like 
	# reshape, transpose, squeeze later
	#------------------------------------
	
	@staticmethod
	def reshape(a: Any, shape: int | tuple[int, ...]) -> np.ndarray:
		try:
			result = np.reshape(a, shape=shape)
		except Exception as e:
			raise excp.InputError(f"can't reshape `a`") from e
		return result
	
	#------------------------------------
	# Complex numbers operations
	#------------------------------------
	
	@staticmethod
	def real(a: Any) -> np.ndarray:
		try:
			result = np.real(a)
		except Exception as e:
			raise excp.InputError(f"can't find real for `a`") from e
		return result
	
	@staticmethod
	def imag(a: Any) -> np.ndarray:
		try:
			result = np.imag(a)
		except Exception as e:
			raise excp.InputError(f"can't find imag for `a`") from e
		return result
	
	@staticmethod
	def angle(a: Any) -> np.ndarray:
		try:
			result = np.angle(a)
		except Exception as e:
			raise excp.InputError(f"can't find angle for `a`") from e
		return result
	
	#------------------------------------
	# Comparison
	#------------------------------------
	
	@staticmethod
	def lt(a: Any, b: Any) -> np.ndarray:
		try:
			mask = a < b
		except Exception as e:
			raise excp.InputError(f"`a` and `b` can't be compared") from e
		return mask
	
	@staticmethod
	def le(a: Any, b: Any) -> np.ndarray:
		try:
			mask = a <= b
		except Exception as e:
			raise excp.InputError(f"`a` and `b` can't be compared") from e
		return mask
	
	@staticmethod
	def gt(a: Any, b: Any) -> np.ndarray:
		try:
			mask = a > b
		except Exception as e:
			raise excp.InputError(f"`a` and `b` can't be compared") from e
		return mask
	
	@staticmethod
	def ge(a: Any, b: Any) -> np.ndarray:
		try:
			mask = a >= b
		except Exception as e:
			raise excp.InputError(f"`a` and `b` can't be compared") from e
		return mask
	
	@staticmethod
	def eq(a: Any, b: Any) -> np.ndarray:
		try:
			mask = a == b
		except Exception as e:
			raise excp.InputError(f"`a` and `b` can't be compared") from e
		return mask
	
	@staticmethod
	def ne(a: Any, b: Any) -> np.ndarray:
		try:
			mask = a != b
		except Exception as e:
			raise excp.InputError(f"`a` and `b` can't be compared") from e
		return mask


#!/usr/bin/env python3


import soundfile as sf
from scipy.signal import resample
import numpy as np
from pathlib import Path
import tempfile
from scipy.signal import resample
from .youtube_downloader import download
from .audio_converter import convert

def load(path, sr=None, trim=None, mono=True):
	"""
	Loads audio file from various sources.

	.. code-block:: python
		
		import modusa as ms
		audio_fp = ms.load(
			"https://www.youtube.com/watch?v=lIpw9-Y_N0g",
			sr = None, trim=(5, 10))

	Parameters
	----------
	path: str
		- Path to the audio file.
		- YouTube URL.
	sr: int | None
		- Sampling rate to load the audio in.
	trim: number | tuple[number, number] | None
		- Segment of the audio to load.
		- Example: 10 => First 10 seconds, (5, 10) => 5 to 10 seconds.
		- Default: None => Entire audio.
	mono: bool
		- If True, loads the signal in mono.

	Return
	------
	np.ndarray
		- Audio signal.
	int
		- Sampling rate of the loaded audio signal.
	title
		- Title of the loaded audio.
		- Filename without extension or YouTube title.
	"""
	# Check if the path is YouTube
	if ".youtube." in str(path):
		# Download the audio in temp directory using tempfile module
		with tempfile.TemporaryDirectory() as tmpdir:
			# Download
			audio_fp: Path = download(url=path, content_type="audio", output_dir=Path(tmpdir))
			
			# Convert the audio to ".wav" form for loading
			wav_audio_fp: Path = convert(inp_audio_fp=audio_fp, output_audio_fp=audio_fp.with_suffix(".wav"))
			
			# Load the audio in memory
			audio_data, audio_sr = sf.read(wav_audio_fp)
			title = audio_fp.stem
	else:
		# Check if the file exists
		fp = Path(path)
		
		if not fp.exists():
			raise FileNotFoundError(f"{path} does not exist.")
			
		# Load the audio in memory
		audio_data, audio_sr = sf.read(fp)
		title = fp.stem
		
	# Convert to mono if requested and it's multi-channel
	if mono and audio_data.ndim > 1:
		audio_data = audio_data.mean(axis=1)
		
	# Resample if needed
	if sr is not None and audio_sr != sr:
		n_samples = int(len(audio_data) * sr / audio_sr)
		
		if audio_data.ndim == 1:
			# Mono
			audio_data = resample(audio_data, n_samples)
		else:
			# Stereo or multi-channel: resample each channel independently
			audio_data = np.stack([
				resample(audio_data[:, ch], n_samples)
				for ch in range(audio_data.shape[1])
			], axis=1)
			
		audio_sr = sr
		
	# Trim if requested
	if trim is not None:
		if isinstance(trim, (int, float)):
			trim = (0, trim)
		elif isinstance(trim, tuple) and len(trim) > 1:
			trim = (trim[0], trim[1])
		else:
			raise ValueError(f"Invalid trim type or length: {type(trim)}, len={len(trim)}")
			
		start = int(trim[0] * audio_sr)
		end = int(trim[1] * audio_sr)
		audio_data = audio_data[start:end]
		
	# Clip to avoid out-of-range playback issues
	if np.issubdtype(audio_data.dtype, np.floating):
		audio_data = np.clip(audio_data, -1.0, 1.0)
		
	return audio_data.T, audio_sr, title


#!/usr/bin/env python3


#----------------------------------------
# Base class errors
#----------------------------------------
class ModusaBaseError(Exception):
	"""
	Ultimate base class for any kind of custom errors.
	"""
	pass

class TypeError(ModusaBaseError):
	pass
	
class InputError(ModusaBaseError):
	"""
	Any Input type error.
	"""

class InputTypeError(ModusaBaseError):
	"""
	Any Input type error.
	"""

class InputValueError(ModusaBaseError):
	"""
	Any Input type error.
	"""
	
class OperationNotPossibleError(ModusaBaseError):
	"""
	Any errors if there is an operations
	failure.
	"""
	
class ImmutableAttributeError(ModusaBaseError):
	"""Raised when attempting to modify an immutable attribute."""
	pass
	
class FileNotFoundError(ModusaBaseError):
	"""Raised when file does not exist."""
	pass
	
class PluginInputError(ModusaBaseError):
	pass

class PluginOutputError(ModusaBaseError):
	pass


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import immutable_property, validate_args_type
from modusa.signals.base import ?
from typing import Self, Any
import numpy as np

class {class_name}():
	"""

	"""
	
	#--------Meta Information----------
	_name = ""
	_nickname = "" # This is to be used in repr/str methods
	_description = ""
	_author_name = "{author_name}"
	_author_email = "{author_email}"
	_created_at = "{date_created}"
	#----------------------------------
	
	@validate_args_type()
	def __init__(self):
		super().__init__() # Instantiating `ModusaSignal` class
	
		self.title = "" # This title will be used as plot title by default
		
	#-----------------------------------
	# Properties
	#-----------------------------------
		
		
	#===================================
		
		
		
		
		
		
	#-----------------------------------
	# Tools
	#-----------------------------------
		
		
	#===================================


#!/usr/bin/env python3

from datetime import date
from pathlib import Path
import questionary
import sys

ROOT_DIR = Path(__file__).parents[3].resolve()
SRC_CODE_DIR = ROOT_DIR / "src/modusa"
TESTS_DIR = ROOT_DIR / "tests"
TEMPLATES_DIR = ROOT_DIR / "src/modusa/devtools/templates"

PLUGIN_INFO = {
	"template_fp": TEMPLATES_DIR / "plugin.py",
	"test_template_fp": TEMPLATES_DIR / "test.py",
	"template_dump_dp": SRC_CODE_DIR / "plugins",
	"test_template_dump_dp": TESTS_DIR / "test_plugins"
}
IO_INFO = {
	"template_fp": TEMPLATES_DIR / "io.py",
	"test_template_fp": TEMPLATES_DIR / "test.py",
	"template_dump_dp": SRC_CODE_DIR / "io",
	"test_template_dump_dp": TESTS_DIR / "test_io"
}
GENERATOR_INFO = {
	"template_fp": TEMPLATES_DIR / "generator.py",
	"test_template_fp": TEMPLATES_DIR / "test.py",
	"template_dump_dp": SRC_CODE_DIR / "generators",
	"test_template_dump_dp": TESTS_DIR / "test_generators"
}
MODEL_INFO = {
	"template_fp": TEMPLATES_DIR / "model.py",
	"test_template_fp": TEMPLATES_DIR / "test.py",
	"template_dump_dp": SRC_CODE_DIR / "models",
	"test_template_dump_dp": TESTS_DIR / "test_models"
}
TOOL_INFO = {
	"template_fp": TEMPLATES_DIR / "tool.py",
	"test_template_fp": TEMPLATES_DIR / "test.py",
	"template_dump_dp": SRC_CODE_DIR / "tools",
	"test_template_dump_dp": TESTS_DIR / "test_tools"
}


class TemplateGenerator():
	"""
	Generates template for `plugin`, `engine`, `signal`, `generator` along with its corresponding `test` file
	in the `tests` directory.
	"""
	
	
	@staticmethod
	def get_path_info(for_what: str):
		if for_what == "plugin": return PLUGIN_INFO
		if for_what == "io": return IO_INFO
		if for_what == "model": return MODEL_INFO
		if for_what == "tool": return TOOL_INFO
		if for_what == "generator": return GENERATOR_INFO
	
	@staticmethod
	def ask_questions(for_what: str, path_info: dict) -> dict:
		"""Asks question about the template to be generated."""
		print("----------------------")
		print(for_what.upper())
		print("----------------------")
		module_name = questionary.text("Module name (snake_case): ").ask()
		
		if module_name is None:
			sys.exit(1)
			
		if not module_name.endswith(".py"): # Adding extension
			module_name = module_name + ".py"
		# Checking if the module name already exists in the dump directory
		if (path_info["template_dump_dp"] / module_name).exists():
			print(f"⚠️ File already exists, choose another name.")
			sys.exit(1)
		
		class_name = questionary.text(f"Class name (CamelCase): ").ask()
		if class_name is None:
			sys.exit(1)
			
		author_name = questionary.text("Author name: ").ask()
		if author_name is None:
			sys.exit(1)
		
		author_email = questionary.text("Author email: ").ask()
		if author_email is None:
			sys.exit(1)
		
		answers = {"for_what": for_what, "module_name": module_name, "class_name": class_name, "author_name": author_name, "author_email": author_email, "date_created": date.today()}
			
		return answers
	
	@staticmethod
	def load_template_file(template_fp: Path) -> str:
		"""Loads template file."""
		if not template_fp.exists():
			print(f"❌ Template not found: {template_fp}")
			sys.exit(1)
		
		template_code = template_fp.read_text()
		
		return template_code
	
	@staticmethod
	def fill_placeholders(template_code: str, placehoders_dict: dict) -> str:
		"""Fills placeholder in the template with the user input from CLI."""
		template_code = template_code.format(**placehoders_dict)  # Fill placeholders
		return template_code
	
	@staticmethod
	def save_file(content: str, output_path: Path) -> None:
		"""Saves file in the correct directory with the right tempalate content."""
		output_path.parent.mkdir(parents=True, exist_ok=True)
		output_path.write_text(content)
		
		# Generating a corresponding test file too
		what_for = output_path.parent.name # plugins, generators, ...
		module_name = output_path.name # this.py
	
	
	@staticmethod
	def create_template(for_what: str) -> None:
		
		# Load correct path location info for the templates and where to dump the files
		path_info: dict = TemplateGenerator.get_path_info(for_what)
		
		# Ask basic questions to create the template for `plugin`, `generator`, ...
		answers: dict = TemplateGenerator.ask_questions(for_what, path_info)
		
		# Load the correct template file and test file
		template_code: str = TemplateGenerator.load_template_file(template_fp=path_info['template_fp'])
		test_code: str = TemplateGenerator.load_template_file(template_fp=path_info['test_template_fp'])
		
		# Update the dynamic values based on the answers
		template_code: str = TemplateGenerator.fill_placeholders(template_code, answers)
		test_code: str = TemplateGenerator.fill_placeholders(test_code, answers)
		
		# Save it to a file and put it in the correct folder
		TemplateGenerator.save_file(content=template_code, output_path=path_info['template_dump_dp'] / answers['module_name'])
		TemplateGenerator.save_file(content=test_code, output_path=path_info['test_template_dump_dp'] / f"test_{answers['module_name']}")
		
		print(f"✅ {for_what}:", "open " + str(path_info['template_dump_dp'] / answers['module_name']))
		print(f"✅ test:", "open " + str(path_info['test_template_dump_dp'] / f"test_{answers['module_name']}"))


#!/usr/bin/env python3


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.patches import Rectangle
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

# Helper for 2D plot
def _calculate_extent(x, y):
	# Handle spacing safely
	if len(x) > 1:
		dx = x[1] - x[0]
	else:
		dx = 1  # Default spacing for single value
	if len(y) > 1:
		dy = y[1] - y[0]
	else:
		dy = 1  # Default spacing for single value
		
	return [
		x[0] - dx / 2,
		x[-1] + dx / 2,
		y[0] - dy / 2,
		y[-1] + dy / 2
	]
	
# Helper to load fonts (devnagri)
def set_default_hindi_font():
	"""
	Hindi fonts works for both english and hindi.
	"""
	from pathlib import Path
	import matplotlib as mpl
	import matplotlib.font_manager as fm
	# Path to your bundled font
	font_path = Path(__file__).resolve().parents[1] / "fonts" / "NotoSansDevanagari-Regular.ttf"
	
	# Register the font with matplotlib
	fm.fontManager.addfont(str(font_path))
	
	# Get the font family name from the file
	hindi_font = fm.FontProperties(fname=str(font_path))
	
	# Set as default rcParam
	mpl.rcParams['font.family'] = hindi_font.get_name()

#set_default_hindi_font()

#======== 1D ===========
def plot1d(*args, ann=None, events=None, xlim=None, ylim=None, xlabel=None, ylabel=None, title=None, legend=None, fmt=None, show_grid=False, show_stem=False):
		"""
		Plots a 1D signal using matplotlib.

		.. code-block:: python
	
			import modusa as ms
			import numpy as np
			
			x = np.arange(100) / 100
			y = np.sin(x)
			
			display(ms.plot1d(y, x))
			
	
		Parameters
		----------
		*args : tuple[array-like, array-like] | tuple[array-like]
			- The signal y and axis x to be plotted.
			- If only values are provided, we generate the axis using arange.
			- E.g. (y1, x1), (y2, x2), ...
		ann : list[tuple[Number, Number, str] | None
			- A list of annotations to mark specific points. Each tuple should be of the form (start, end, label).
			- Default: None => No annotation.
		events : list[Number] | None
			- A list of x-values where vertical lines (event markers) will be drawn.
			- Default: None
		xlim : tuple[Number, Number] | None
			- Limits for the x-axis as (xmin, xmax).
			- Default: None
		ylim : tuple[Number, Number] | None
			- Limits for the y-axis as (ymin, ymax).
			- Default: None
		xlabel : str | None
			- Label for the x-axis.
			- - Default: None
		ylabel : str | None
			- Label for the y-axis.
			- Default: None
		title : str | None
			- Title of the plot.
			- Default: None
		legend : list[str] | None
			- List of legend labels corresponding to each signal if plotting multiple lines.
			- Default: None
		fmt: list[str] | None
			- linefmt for different line plots.
			- Default: None
		show_grid: bool
			- If you want to show the grid.
			- Default: False
		show_stem: bool:
			- If you want stem plot.
			- Default: False
	
		Returns
		-------
		plt.Figure
			Matplolib figure.
		"""
		for arg in args:
			if len(arg) not in [1, 2]: # 1 if it just provides values, 2 if it provided axis as well
				raise ValueError(f"1D signal needs to have max 2 arrays (y, x) or simply (y, )")
			
		if isinstance(legend, str): legend = (legend, )
		if legend is not None:
			if len(legend) < len(args):
				raise ValueError(f"`legend` should be provided for each signal.")
		
		if isinstance(fmt, str): fmt = [fmt]
		if fmt is not None:
			if len(fmt) < len(args):
				raise ValueError(f"`fmt` should be provided for each signal.")

		colors = plt.get_cmap('tab10').colors

		fig = plt.figure(figsize=(16, 2))
		gs = gridspec.GridSpec(3, 1, height_ratios=[0.2, 0.2, 1])
		
		signal_ax = fig.add_subplot(gs[2, 0])
		annotation_ax = fig.add_subplot(gs[0, 0], sharex=signal_ax)
		events_ax = fig.add_subplot(gs[1, 0])
		
		# Set lim
		if xlim is not None:
			signal_ax.set_xlim(xlim)
		
		if ylim is not None:
			signal_ax.set_ylim(ylim)
			
		# Add signal plot
		for i, signal in enumerate(args):
			if len(signal) == 1:
				y = signal[0]
				x = np.arange(y.size)
				if legend is not None:
					if show_stem is True:
						markerline, stemlines, baseline = signal_ax.stem(x, y, label=legend[i])
						markerline.set_color(colors[i])
						stemlines.set_color(colors[i])
						baseline.set_color("k")
					else:
						if fmt is not None:
							signal_ax.plot(x, y, fmt[i], markersize=4, label=legend[i])
						else:
							signal_ax.plot(x, y, color=colors[i], label=legend[i])
				else:
					if show_stem is True:
						markerline, stemlines, baseline = signal_ax.stem(x, y)
						markerline.set_color(colors[i])
						stemlines.set_color(colors[i])
						baseline.set_color("k")
					else:
						if fmt is not None:
							signal_ax.plot(x, y, fmt[i], markersize=4)
						else:
							signal_ax.plot(x, y, color=colors[i])
						
			elif len(signal) == 2:
				y, x = signal[0], signal[1]
				if legend is not None:
					if show_stem is True:
						markerline, stemlines, baseline = signal_ax.stem(x, y, label=legend[i])
						markerline.set_color(colors[i])
						stemlines.set_color(colors[i])
						baseline.set_color("k")
					else:
						if fmt is not None:
							signal_ax.plot(x, y, fmt[i], markersize=4, label=legend[i])
						else:
							signal_ax.plot(x, y, color=colors[i], label=legend[i])
				else:
					if show_stem is True:
						markerline, stemlines, baseline = signal_ax.stem(x, y)
						markerline.set_color(colors[i])
						stemlines.set_color(colors[i])
						baseline.set_color("k")
					else:
						if fmt is not None:
							signal_ax.plot(x, y, fmt[i], markersize=4)
						else:
							signal_ax.plot(x, y, color=colors[i])
							
		
		# Add annotations
		if ann is not None:
			annotation_ax.set_ylim(0, 1) # For consistent layout
			# Determine visible x-range
			x_view_min = xlim[0] if xlim is not None else np.min(x)
			x_view_max = xlim[1] if xlim is not None else np.max(x)
			
			for i, (start, end, tag) in enumerate(ann):
				# We make sure that we only plot annotation that are within the x range of the current view
				if start >= x_view_max or end <= x_view_min:
					continue
				
				# Clip boundaries to xlim
				start = max(start, x_view_min)
				end = min(end, x_view_max)
				
				box_colors = ["gray", "lightgray"] # Alternates color between two
				box_color = box_colors[i % 2]
				
				width = end - start
				rect = Rectangle((start, 0), width, 1, color=box_color, alpha=0.7)
				annotation_ax.add_patch(rect)
				
				text_obj = annotation_ax.text(
					(start + end) / 2, 0.5, tag,
					ha='center', va='center',
					fontsize=10, color="black", fontweight='bold', zorder=10, clip_on=True
				)
				
				text_obj.set_clip_path(rect)
				
				
		# Add vlines
		if events is not None:
#			if not isinstance(events, tuple):
#				raise TypeError(f"`events` should be tuple, got {type(events)}")
			
			for xpos in events:
				if xlim is not None:
					if xlim[0] <= xpos <= xlim[1]:
						annotation_ax.axvline(x=xpos, color='red', linestyle='-', linewidth=1.5)
				else:
					annotation_ax.axvline(x=xpos, color='red', linestyle='-', linewidth=1.5)
					
		# Add legend
		if legend is not None:
			handles, labels = signal_ax.get_legend_handles_labels()
			fig.legend(handles, labels, loc='upper right', bbox_to_anchor=(0.9, 1.2), ncol=len(legend), frameon=True)
			
		# Set title, labels
		if title is not None:
			annotation_ax.set_title(title, pad=10, size=11)
		if xlabel is not None:
			signal_ax.set_xlabel(xlabel)
		if ylabel is not None:
			signal_ax.set_ylabel(ylabel)
			
		# Add grid to the plot
		if show_grid is True:
			signal_ax.grid(True, linestyle=':', linewidth=0.7, color='gray', alpha=0.7)
		
		# Remove the boundaries and ticks from an axis
		if ann is not None:
			annotation_ax.tick_params(left=False, bottom=False, labelleft=False, labelbottom=False)
		else:
			annotation_ax.axis("off")
			
		if events is not None:
			events_ax.tick_params(left=False, bottom=False, labelleft=False, labelbottom=False)
		else:
			events_ax.axis("off")
		
		
		fig.subplots_adjust(hspace=0.01, wspace=0.05)
		plt.close()
		return fig

#======== 2D ===========
def plot2d(*args, ann=None, events=None, xlim=None, ylim=None, origin="lower", Mlabel=None, xlabel=None, ylabel=None, title=None, legend=None, lm=False, show_grid=False):
	"""
	Plots a 2D matrix (e.g., spectrogram or heatmap) with optional annotations and events.

	.. code-block:: python

		import modusa as ms
		import numpy as np
		
		M = np.random.random((10, 30))
		y = np.arange(M.shape[0])
		x = np.arange(M.shape[1])
		
		display(ms.plot2d(M, y, x))

	Parameters
	----------
	*args : tuple[array-like, array-like]
		- The signal values to be plotted.
		- E.g. (M1, y1, x1), (M2, y2, x2), ...
	ann : list[tuple[Number, Number, str]] | None
		- A list of annotation spans. Each tuple should be (start, end, label).
		- Default: None (no annotations).
	events : list[Number] | None
		- X-values where vertical event lines will be drawn.
		- Default: None.
	xlim : tuple[Number, Number] | None
		- Limits for the x-axis as (xmin, xmax).
		- Default: None (auto-scaled).
	ylim : tuple[Number, Number] | None
		- Limits for the y-axis as (ymin, ymax).
		- Default: None (auto-scaled).
	origin : {'upper', 'lower'}
		- Origin position for the image display. Used in `imshow`.
		- Default: "lower".
	Mlabel : str | None
		- Label for the colorbar (e.g., "Magnitude", "Energy").
		- Default: None.
	xlabel : str | None
		- Label for the x-axis.
		- Default: None.
	ylabel : str | None
		- Label for the y-axis.
		- Default: None.
	title : str | None
		- Title of the plot.
		- Default: None.
	legend : list[str] | None
		- Legend labels for any overlaid lines or annotations.
		- Default: None.
	lm: bool
		- Adds a circular marker for the line.
		- Default: False
		- Useful to show the data points.
	show_grid: bool
		- If you want to show the grid.
		- Default: False

	Returns
	-------
	matplotlib.figure.Figure
		The matplotlib Figure object.
	"""
	
	for arg in args:
		if len(arg) not in [1, 2, 3]: # Either provide just the matrix or with both axes info
			raise ValueError(f"Data to plot needs to have 3 arrays (M, y, x)")
	if isinstance(legend, str): legend = (legend, )
	
	fig = plt.figure(figsize=(16, 4))
	gs = gridspec.GridSpec(3, 1, height_ratios=[0.2, 0.1, 1]) # colorbar, annotation, signal

	colors = plt.get_cmap('tab10').colors
	
	signal_ax = fig.add_subplot(gs[2, 0])
	annotation_ax = fig.add_subplot(gs[1, 0], sharex=signal_ax)
	
	colorbar_ax = fig.add_subplot(gs[0, 0])
	colorbar_ax.axis("off")
	
	
	# Add lim
	if xlim is not None:
		signal_ax.set_xlim(xlim)
		
	if ylim is not None:
		signal_ax.set_ylim(ylim)
		
	# Add signal plot
	i = 0 # This is to track the legend for 1D plots
	for signal in args:
		
		data = signal[0] # This can be 1D or 2D (1D meaning we have to overlay on the matrix)
			
		if data.ndim == 1: # 1D
			if len(signal) == 1: # It means that the axis was not passed
				x = np.arange(data.shape[0])
			else:
				x = signal[1]
			
			if lm is False:
				if legend is not None:
					signal_ax.plot(x, data, label=legend[i])
					signal_ax.legend(loc="upper right")
				else:
					signal_ax.plot(x, data)
			else:
				if legend is not None:
					signal_ax.plot(x, data, marker="o", markersize=7, markerfacecolor='red', linestyle="--", linewidth=2, label=legend[i])
					signal_ax.legend(loc="upper right")
				else:
					signal_ax.plot(x, data, marker="o", markersize=7, markerfacecolor='red', linestyle="--", linewidth=2)
					
			i += 1
			
		elif data.ndim == 2: # 2D
			M = data
			if len(signal) == 1: # It means that the axes were not passed
				y = np.arange(M.shape[0])
				x = np.arange(M.shape[1])
				extent = _calculate_extent(x, y)
				im = signal_ax.imshow(M, aspect="auto", origin=origin, cmap="gray_r", extent=extent)
				
			elif len(signal) == 3: # It means that the axes were passed
				M, y, x = signal[0], signal[1], signal[2]
				extent = _calculate_extent(x, y)
				im = signal_ax.imshow(M, aspect="auto", origin=origin, cmap="gray_r", extent=extent)
	
	# Add annotations
	if ann is not None:
		annotation_ax.set_ylim(0, 1) # For consistent layout
		# Determine visible x-range
		x_view_min = xlim[0] if xlim is not None else np.min(x)
		x_view_max = xlim[1] if xlim is not None else np.max(x)
		
		for i, (start, end, tag) in enumerate(ann):
			# We make sure that we only plot annotation that are within the x range of the current view
			if start >= x_view_max or end <= x_view_min:
				continue
			
			# Clip boundaries to xlim
			start = max(start, x_view_min)
			end = min(end, x_view_max)
				
			color = colors[i % len(colors)]
			width = end - start
			rect = Rectangle((start, 0), width, 1, color=color, alpha=0.7)
			annotation_ax.add_patch(rect)
			text_obj = annotation_ax.text(
				(start + end) / 2, 0.5, tag,
				ha='center', va='center',
				fontsize=10, color='white', fontweight='bold', zorder=10, clip_on=True
			)
			
			text_obj.set_clip_path(rect)
			
	# Add vlines
	if events is not None:
		for xpos in events:
			if xlim is not None:
				if xlim[0] <= xpos <= xlim[1]:
					annotation_ax.axvline(x=xpos, color='black', linestyle='--', linewidth=1.5)
			else:
				annotation_ax.axvline(x=xpos, color='black', linestyle='--', linewidth=1.5)
	
	# Add legend incase there are 1D overlays
	if legend is not None:
		handles, labels = signal_ax.get_legend_handles_labels()
		if handles:  # Only add legend if there's something to show
			signal_ax.legend(handles, labels, loc="upper right")
	
	# Add colorbar
	# Create an inset axis on top-right of signal_ax
	cax = inset_axes(
		colorbar_ax,
		width="20%",      # percentage of parent width
		height="20%",      # height in percentage of parent height
		loc='upper right',
		bbox_to_anchor=(0, 0, 1, 1),
		bbox_transform=colorbar_ax.transAxes,
		borderpad=1
	)
	
	cbar = plt.colorbar(im, cax=cax, orientation='horizontal')
	cbar.ax.xaxis.set_ticks_position('top')
	
	if Mlabel is not None:
		cbar.set_label(Mlabel, labelpad=5)
	
		
	# Set title, labels
	if title is not None:
		annotation_ax.set_title(title, pad=10, size=11)
	if xlabel is not None:
		signal_ax.set_xlabel(xlabel)
	if ylabel is not None:
		signal_ax.set_ylabel(ylabel)
		
	# Add grid to the plot
	if show_grid is True:
		signal_ax.grid(True, linestyle=':', linewidth=0.7, color='gray', alpha=0.7)
	
	# Making annotation axis spines thicker
	if ann is not None:
		annotation_ax.tick_params(left=False, bottom=False, labelleft=False, labelbottom=False)
	else:
		annotation_ax.axis("off")

	fig.subplots_adjust(hspace=0.01, wspace=0.05)
	plt.close()
	return fig

#======== Plot distribution ===========
def plot_dist(*args, ann=None, xlim=None, ylim=None, ylabel=None, xlabel=None, title=None, legend=None, show_hist=True, npoints=200, bins=30):
		"""
		Plot distribution.

		.. code-block:: python
			
			import modusa as ms
			import numpy as np
			np.random.seed(42)
			data = np.random.normal(loc=1, scale=1, size=1000)
			ms.plot_dist(data, data+5, data-10, ann=[(0, 1, "A")], legend=("D1", "D2", "D3"), ylim=(0, 1), xlabel="X", ylabel="Counts", title="Distribution")

		Parameters
		----------
		*args: ndarray
			- Data arrays for which distribution needs to be plotted.
			- Arrays will be flattened.
		ann : list[tuple[Number, Number, str] | None
			- A list of annotations to mark specific points. Each tuple should be of the form (start, end, label).
			- Default: None => No annotation.
		events : list[Number] | None
			- A list of x-values where vertical lines (event markers) will be drawn.
			- Default: None
		xlim : tuple[Number, Number] | None
			- Limits for the x-axis as (xmin, xmax).
			- Default: None
		ylim : tuple[Number, Number] | None
			- Limits for the y-axis as (ymin, ymax).
			- Default: None
		xlabel : str | None
			- Label for the x-axis.
			- - Default: None
		ylabel : str | None
			- Label for the y-axis.
			- Default: None
		title : str | None
			- Title of the plot.
			- Default: None
		legend : list[str] | None
			- List of legend labels corresponding to each signal if plotting multiple distributions.
			- Default: None
		show_hist: bool
			- Want to show histogram as well.
		npoints: int
			- Number of points for which gaussian needs to be computed between min and max.
			- Higher value means more points are evaluated with the fitted gaussian, thereby higher resolution.
		bins: int
			- The number of bins for histogram.
			- This is used only to plot the histogram.

		Returns
		-------
		plt.Figure
			- Matplotlib figure.
		"""
		from scipy.stats import gaussian_kde
	
		if isinstance(legend, str):
			legend = (legend, )
			
		if legend is not None:
			if len(legend) < len(args):
				raise ValueError(f"Legend should be provided for each signal.")
					
		# Create figure
		fig = plt.figure(figsize=(16, 4))
		gs = gridspec.GridSpec(2, 1, height_ratios=[0.1, 1])
	
		colors = plt.get_cmap('tab10').colors
	
		dist_ax = fig.add_subplot(gs[1, 0])
		annotation_ax = fig.add_subplot(gs[0, 0], sharex=dist_ax)
	
		# Set limits
		if xlim is not None:
			dist_ax.set_xlim(xlim)
			
		if ylim is not None:
			dist_ax.set_ylim(ylim)
			
		# Add plot
		for i, data in enumerate(args):
			# Fit gaussian to the data
			kde = gaussian_kde(data)
		
			# Create points to evaluate KDE
			x = np.linspace(np.min(data), np.max(data), npoints)
			y = kde(x)
		
			if legend is not None:
				dist_ax.plot(x, y, color=colors[i], label=legend[i])
				if show_hist is True:
					dist_ax.hist(data, bins=bins, density=True, alpha=0.3, facecolor=colors[i], edgecolor='black', label=legend[i])
			else:
				dist_ax.plot(x, y, color=colors[i])
				if show_hist is True:
					dist_ax.hist(data, bins=bins, density=True, alpha=0.3, facecolor=colors[i], edgecolor='black')
							
		# Add annotations
		if ann is not None:
			annotation_ax.set_ylim(0, 1) # For consistent layout
			# Determine visible x-range
			x_view_min = xlim[0] if xlim is not None else np.min(x)
			x_view_max = xlim[1] if xlim is not None else np.max(x)
			for i, (start, end, tag) in enumerate(ann):
				# We make sure that we only plot annotation that are within the x range of the current view
				if start >= x_view_max or end <= x_view_min:
					continue
				
				# Clip boundaries to xlim
				start = max(start, x_view_min)
				end = min(end, x_view_max)
					
				color = colors[i % len(colors)]
				width = end - start
				rect = Rectangle((start, 0), width, 1, color=color, alpha=0.7)
				annotation_ax.add_patch(rect)
			
				text_obj = annotation_ax.text((start + end) / 2, 0.5, tag, ha='center', va='center', fontsize=10, color='white', fontweight='bold', zorder=10, clip_on=True)
				text_obj.set_clip_path(rect)
					
		# Add legend
		if legend is not None:
			handles, labels = dist_ax.get_legend_handles_labels()
			fig.legend(handles, labels, loc='upper right', bbox_to_anchor=(0.9, 1.1), ncol=len(legend), frameon=True)
			
		# Set title, labels
		if title is not None:
			annotation_ax.set_title(title, pad=10, size=11)
		if xlabel is not None:
			dist_ax.set_xlabel(xlabel)
		if ylabel is not None:
			dist_ax.set_ylabel(ylabel)
			
		# Remove the boundaries and ticks from annotation axis
		if ann is not None:
			annotation_ax.tick_params(left=False, bottom=False, labelleft=False, labelbottom=False)
		else:
			annotation_ax.axis("off")
			
		fig.subplots_adjust(hspace=0.01, wspace=0.05)
		plt.close()
		return fig


#!/usr/bin/env python3

#---------------------------------
# Author: Ankit Anand
# Date: 22/08/25
# Email: ankit0.anand0@gmail.com
#---------------------------------


import sounddevice as sd
import numpy as np
import ipywidgets as widgets
from IPython.display import display, clear_output, Audio

def record():
	"""
	Create a UI to record audio in jupyter notebook, the 
	recorded signal is available as array.

	.. code-block:: python
		
		import modusa as ms
		result = ms.record()
		y, sr, title = result() # Keep it in the next cell

	Returns
	-------
	Callable
		A lambda function that returns y(audio signal), sr(sampling rate), title(title set in the UI)
	"""
	devices = sd.query_devices()
	device_options = [(d['name'][:30], i) for i, d in enumerate(devices) if d['max_input_channels'] > 0]
	
	# Controls
	device_dropdown = widgets.Dropdown(
		options=device_options,
		description="Microphone:",
		layout=widgets.Layout(width="300px")
	)
	
	sr_dropdown = widgets.Dropdown(
		options=[('16 Khz', 16000), ('22.05 Khz', 22050), ('44.1 Khz', 44100), ('48 Khz', 48000)],
		value=22050,
		description="Sample Rate:",
		layout=widgets.Layout(width="300px")
	)
	
	title_box = widgets.Text(
		placeholder="Title",
		description="Title:",
		layout=widgets.Layout(width="300px")
	)
	
	toggle_button = widgets.Button(
		description="Record",
		button_style="",
	)
	
	status = widgets.HTML(value="")
	out = widgets.Output()
	
	# State
	recording = {"data": None, "sr": None, "title": None}
	stream = {"obj": None, "frames": [], "recording": False}
	
	def callback(indata, frames, time, status):
		if not status:
			stream["frames"].append(indata.copy())
			
	def on_toggle(b):
		if not stream["recording"]:
			stream["frames"].clear()
			sr = sr_dropdown.value
			device_id = device_dropdown.value
			
			stream["obj"] = sd.InputStream(callback=callback, channels=1, samplerate=sr, device=device_id)
			stream["obj"].start()
			stream["recording"] = True
			
			toggle_button.description = "Stop"
			toggle_button.button_style = "danger"
			status.value = "Recording..."
		else:
			stream["obj"].stop()
			stream["obj"].close()
			
			sr = sr_dropdown.value
			y = np.concatenate(stream["frames"], axis=0).flatten()
			title = title_box.value.strip() or "Recording"
			
			recording["data"], recording["sr"], recording["title"] = y, sr, title
			record.result = (y, sr, title)
			stream["recording"] = False
			
			toggle_button.description = "Record"
			toggle_button.button_style = "success"
			
			with out:
				clear_output()
				display(Audio(y, rate=sr))
				
	toggle_button.on_click(on_toggle)
	
	# Layout
	ui = widgets.VBox([
		device_dropdown,
		sr_dropdown,
		title_box,
		widgets.HBox([toggle_button]),
		out
	])
	
	display(ui)
	record.result = None
	return lambda: record.result


#!/usr/bin/env python3

import logging

class PATHS:
	from pathlib import Path
	ROOT_DP: Path = Path(__file__).parents[2].resolve()
	AUDIO_DP: Path = ROOT_DP / "data" / "audio"
	EXAMPLE_AUDIO_FP: Path = AUDIO_DP / "Arko - Nazm Nazm.mp3"
	LABELS_CSV_FP: Path = ROOT_DP / "data" / "label_data.csv"
	REPORTS_DP: Path = ROOT_DP / "data" / "reports"

class DEFAULT_SETTINGS:
	SR: int = 44100
	LOG_LEVEL = logging.WARNING
	
	class STFT:
		N_FFT: int = 2048
		WIN_SIZE: int = 2048
		HOP_SIZE: int = 512
		WINDOW: str = "hann"
	
	class NOVELTY:
		GAMMA: int = 10
		LOCAL_AVG: int = 40


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import immutable_property, validate_args_type, plugin_safety_check
from modusa.plugins.base import ModusaPlugin


class {class_name}(ModusaPlugin):
	"""

	"""
	
	#--------Meta Information----------
	_name = ""
	_description = ""
	_author_name = "{author_name}"
	_author_email = "{author_email}"
	_created_at = "{date_created}"
	#----------------------------------
	
	def __init__(self):
		super().__init__()
		
	@immutable_property(error_msg="Mutation not allowed.")
	def allowed_input_signal_types(self) -> tuple[type, ...]:
		return ()
	
	
	@immutable_property(error_msg="Mutation not allowed.")
	def allowed_output_signal_types(self) -> tuple[type, ...]:
		return ()
	
	
	@plugin_safety_check()
	@validate_args_type()
	def apply(self, signal: "") -> "":
		
		# Run the engine here
		
		return 


#!/usr/bin/env python3

import logging

def setup_logging(log_level: int | None = logging.WARNING):
	logging.basicConfig(
		level=log_level,
		format="%(asctime)s | %(levelname)s | %(name)s | %(funcName)s():%(lineno)d\n> %(message)s\n",
		datefmt='%Y-%m-%d %H:%M:%S'
	)
	# Silence 3rd-party spam
	logging.getLogger("matplotlib").setLevel(logging.WARNING)
	logging.getLogger("PIL").setLevel(logging.WARNING)
	logging.getLogger("numba").setLevel(logging.WARNING)
	logging.getLogger("py").setLevel(logging.ERROR)
	
def get_logger(name=None):
	return logging.getLogger(name)



#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import validate_args_type
from .base import ModusaGenerator
from modusa.models.t_ax import TAx
from modusa.models.audio import Audio
from modusa.models.data import Data
import numpy as np
from pathlib import Path

class AudioGen(ModusaGenerator):
	"""
	Provides user friendly APIs to generate instances of different 
	`AudioSignal` instances.
	"""
	
	#--------Meta Information----------
	_name = ""
	_description = ""
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-27"
	#----------------------------------
	
	@staticmethod
	def from_array(
		y: np.ndarray,
		sr: float | int = 1.0,
		t0: float | int = 0.0,
		y_label: str = "Amplitude",
		t_label: str = "Time (sec)",
		title: str = "Audio Signal"
	) -> Audio:
		"""
		Create `AudioSignal` instance from basic data structures.

		.. code-block:: python

			import modusa as ms
			t = ms.tds.from_array([1, 2, 3])
			print(t)
			t.print_info()
		
		Parameters
		----------
		y: np.ndarray
			- Audio data array.
		sr: float | int
			- Sampling rate.
		t0: float | int
			- Start timestamp.
		y_label: str
			- Y label for the signal.
			- Default: "Y"
		t_label: str
			- T label for the signal.
			- Default: "Time (sec)"
		title: str
			- Title for the signal.
			- Default: "1D Signal"
		Returns
		-------
		Audio
			An instance of Audio.
		"""
		assert isinstance(y, np.ndarray)
		assert isinstance(sr, (int, float)) and isinstance(t0, (int, float))
		assert isinstance(y_label, str) and isinstance(t_label, str) and isinstance(title, str)
		
		assert y.ndim == 1
		
		sr = float(sr)
		t0 = float(t0)
		
		t = TAx(n_points=y.shape[0], sr=sr, t0=t0, label=t_label) # Creating a signal axis instance
		y = Data(values=y, label=y_label)
		
		return Audio(y=y, t=t, title=title)
	
	@classmethod
	def from_youtube(cls, url: str, sr: int | float = None):
		"""
		Loads audio from youtube at a given sr.
		The audio is deleted from the device
		after loading.
		
		.. code-block:: python
			
			import modusa as ms
			audio = ms.audio.from_youtube(
				url="https://www.youtube.com/watch?v=lIpw9-Y_N0g", 
				sr=None
			)

		PARAMETERS
		----------
		url: str
			Link to the YouTube video.
		sr: int
			Sampling rate to load the audio in.
		
		Returns
		-------
		Audio:
			An `Audio` instance with loaded audio content from YouTube.
		"""
		
		from modusa.tools.youtube_downloader import YoutubeDownloader
		from modusa import convert
		import soundfile as sf
		from scipy.signal import resample
		import tempfile
		
		# Download the audio in temp directory using tempfile module
		with tempfile.TemporaryDirectory() as tmpdir:
			audio_fp: Path = YoutubeDownloader.download(url=url, content_type="audio", output_dir=Path(tmpdir))
			
			# Convert the audio to ".wav" form for loading
			wav_audio_fp: Path = convert(inp_audio_fp=audio_fp, output_audio_fp=audio_fp.with_suffix(".wav"))
			
			# Load the audio in memory
			audio_data, audio_sr = sf.read(wav_audio_fp)
			
			# Convert to mono if it's multi-channel
			if audio_data.ndim > 1:
				audio_data = audio_data.mean(axis=1)
				
			# Resample if needed
			if sr is not None:
				if audio_sr != sr:
					n_samples = int(len(audio_data) * sr / audio_sr)
					audio_data = resample(audio_data, n_samples)
					audio_sr = sr
					
		audio = cls.from_array(y=audio_data, sr=audio_sr, title=audio_fp.stem)
		
		return audio
	
	@classmethod
	def from_filepath(cls, fp: str | Path, sr: int | float = None):
		"""
		Loads audio from filepath at a given sr.
		
		.. code-block:: python
			
			import modusa as ms
			audio = ms.audio.from_filepath(
				fp="path/to/audio.wav",
				sr=None
			)

		PARAMETERS
		----------
		fp: str | Path
			Audio file path.
		sr: int
			Sampling rate to load the audio in.
		
		Returns
		-------
		Audio:
			An `Audio` instance with loaded audio content.
		"""
		import soundfile as sf
		from scipy.signal import resample
		from pathlib import Path
		
		fp = Path(fp)
		# Load the audio in memory
		audio_data, audio_sr = sf.read(fp)
		
		# Convert to mono if it's multi-channel
		if audio_data.ndim > 1:
			audio_data = audio_data.mean(axis=1)
			
		# Resample if needed
		if sr is not None:
			if audio_sr != sr:
				n_samples = int(len(audio_data) * sr / audio_sr)
				audio_data = resample(audio_data, n_samples)
				audio_sr = sr
				
		audio = cls.from_array(y=audio_data, sr=audio_sr, title=fp.stem)
		
		return audio
	


#!/usr/bin/env python3


from .generate_template import TemplateGenerator
from .generate_docs_source import generate_docs_source
from .list_plugins import list_plugins
from . import list_authors

import argparse
import sys

def main():
	try:
		parser = argparse.ArgumentParser(
			prog="modusa-dev",
			description="Modusa CLI Tools"
		)
		subparsers = parser.add_subparsers(dest="group", required=True)
		
		# --- CREATE group ---
		create_parser = subparsers.add_parser("create", help="Create new Modusa components")
		create_subparsers = create_parser.add_subparsers(dest="what", required=True)
		
		create_subparsers.add_parser("tool", help="Create a new tool class").set_defaults(func=lambda:TemplateGenerator.create_template("tool"))
		create_subparsers.add_parser("plugin", help="Create a new plugin class").set_defaults(func=lambda:TemplateGenerator.create_template("plugin"))
		create_subparsers.add_parser("model", help="Create a new model class").set_defaults(func=lambda:TemplateGenerator.create_template("model"))
		create_subparsers.add_parser("generator", help="Create a new signal generator class").set_defaults(func=lambda:TemplateGenerator.create_template("generator"))
		create_subparsers.add_parser("io", help="Create a new IO class").set_defaults(func=lambda:TemplateGenerator.create_template("io"))
		create_subparsers.add_parser("docs", help="Generate the docs").set_defaults(func=lambda:generate_docs_source())
		
		# --- LIST group ---
		list_parser = subparsers.add_parser("list", help="List information about Modusa components")
		list_subparsers = list_parser.add_subparsers(dest="what", required=True)
		
		list_subparsers.add_parser("plugins", help="List available plugins").set_defaults(func=list_plugins)
		list_subparsers.add_parser("authors", help="List plugin authors").set_defaults(func=list_authors)
		
		# --- Parse and execute ---
		args = parser.parse_args()
		args.func()
		
	except KeyboardInterrupt:
		print("\n❌ Aborted by user.")
		sys.exit(1)
		


#!/usr/bin/env python3

#---------------------------------
# Author: Ankit Anand
# Date: 13/09/25
# Email: ankit0.anand0@gmail.com
#---------------------------------


def save(y, sr, out_path):
	"""
	Saves an array as an audio file.
	
	Parameters
	----------
	y: ndarray
		- Audio signal array.
	sr: number [> 0]
		- Sampling rate of the audio signal.
	out_path: str
		- Output path of the audio file.
	
	Returns
	-------
	None
	"""
	
	import soundfile as sf
	
	sf.write(file=out_path, data=y, samplerate=sr)



#!/usr/bin/env python3

from modusa import excp

#--------Meta Information----------
_class_name = "{class_name}"
_author_name = "{author_name}"
_author_email = "{author_email}"
_created_at = "{date_created}"
#----------------------------------


#!/usr/bin/env python3

from modusa.decorators import plugin_safety_check
from abc import ABC, abstractmethod
from typing import Any

class ModusaPlugin(ABC):
	"""
	Base class for any Plugin that follows the modusa framework.
	"""
	
	@property
	@abstractmethod
	def allowed_input_signal_types(self) -> tuple[type, ...]:
		"""
		Define the expected signal types for this plugin.
		
		Note
		----
		- Must be implemented by every `ModusaPlugin` subclass.
		- Clearly states what signal types are accepted in `.apply()`.
		- Return a tuple of accepted signal classes.
		
		Examples
		--------
		.. code-block:: python

			# Single type
			from modusa.signals import Signal1D
			return (Signal1D, )
		
			# Multiple types
			from modusa.signals import Signal1D, Signal2D
			return (Signal1D, Signal2D)
		"""
		pass
		
	@property
	@abstractmethod
	def allowed_output_signal_types(self) -> tuple[type, ...]:
		"""
		Defines the allowed return types from the `.apply()` method.
		
		Note
		----
		- Must be implemented by every `ModusaPlugin` subclass.
		- Clearly declares what return types are valid.
		- Return a tuple of accepted types (usually signal classes).
		
		Examples
		--------
		.. code-block:: python

			# Single type allowed
			from modusa.signals import Signal1D
			return (Signal1D, )
		
			# Multiple types allowed
			from modusa.signals import Signal1D, Signal2D
			return (Signal1D, Signal2D)
		
			# Return type can also be None (e.g., for plot-only plugins)
			from modusa.signals import Signal1D, Signal2D
			return (Signal1D, Signal2D, type(None))
		"""
		pass
		
	@plugin_safety_check()
	@abstractmethod
	def apply(self, signal: Any) -> Any:
		"""
		Defines the main processing logic of the plugin.
		
		Note
		----
		- Must be implemented by every `ModusaPlugin` subclass.
		- It is highly advised to wrap it with `@plugin_safety_check()` to:
			- Validate input and output types.
			- Enforce plugin contracts for safe execution.
		
		Warning
		-------
		- You should not make this method as `classmethod` or `staticmethod`, this will break plugin safety check.
		
		Example
		-------
		.. code-block:: python
			
			@plugin_safety_check()
			def apply(self, signal: "TimeDomainSignal") -> "TimeDomainSignal":
				from modusa.engines import SomeEngine
				new_signal: TimeDomainSignal = SomeEngine.run(signal)
				
				return new_signal
		"""
		pass
		
		
	
		


from modusa.utils import excp, config

#=====Giving access to plot functions to plot multiple signals.=====
from modusa.tools import dist_plot, hill_plot, plot, fig
#=====

from modusa.tools import play, convert, record, save
from modusa.tools import download
from modusa.tools import load, load_ann

__version__ = "0.4.21"



#!/usr/bin/env python3

import inspect
import importlib
import pkgutil
from pathlib import Path
from collections import defaultdict

# === Configuration ===
BASE_MODULES = [
    'modusa.tools',
#   'modusa.models',
#   'modusa.generators',
#   'modusa.plugins',
]
OUTPUT_DIRS = [
    Path('docs/source/tools'),
#   Path('docs/source/models'),
#   Path('docs/source/generators'),
#   Path('docs/source/plugins'),
]

# Ensure output directories exist
for output_dir in OUTPUT_DIRS:
    output_dir.mkdir(parents=True, exist_ok=True)

# === Utils ===
def get_classes_grouped_by_module(base_module):
    """
    Returns a dictionary: { module_path: [class_name, ...] }
    """
    found = defaultdict(list)
    module = importlib.import_module(base_module)

    for _, modname, _ in pkgutil.walk_packages(module.__path__, base_module + "."):
        try:
            submodule = importlib.import_module(modname)
            for name, obj in inspect.getmembers(submodule, inspect.isclass):
                if obj.__module__ == modname:
                    found[modname].append(name)
        except Exception as e:
            print(f"⚠️ Skipping {modname} due to import error: {e}")
    return found

def write_module_rst_file(module_path, class_names, output_dir):
    """
    Writes a .rst file for a module, documenting all its classes.
    Filename is based on the module, but title is the first class (Modusa* gets priority).
    """
    filename = output_dir / f"{module_path.split('.')[-1]}.rst"

    # Prioritize 'Modusa*' classes first, then alphabetical
    sorted_classes = sorted(class_names, key=lambda x: (not x.startswith("Modusa"), x.lower()))
    title = sorted_classes[0] if sorted_classes else module_path.split('.')[-1]

    with open(filename, 'w') as f:
        f.write(f"{title}\n{'=' * len(title)}\n\n")
        for class_name in sorted_classes:
            f.write(f".. autoclass:: {module_path}.{class_name}\n")
            f.write("   :members:\n")
            f.write("   :undoc-members:\n")
            f.write("   :show-inheritance:\n\n")
    
    return filename.name

def write_index_rst_file(tools_by_module, output_dir, section_name="Tools"):
    """
    Writes index.rst in the given output_dir with 'base' on top, then other files alphabetically.
    """
    index_file = output_dir / "index.rst"
    with open(index_file, "w") as f:
        f.write(f"{section_name}\n{'=' * len(section_name)}\n\n")
        f.write(".. toctree::\n   :maxdepth: 1\n\n")

        filenames = [module_path.split('.')[-1] for module_path in tools_by_module]
        sorted_filenames = sorted(filenames, key=lambda x: (x != "base", x.lower()))

        for name in sorted_filenames:
            f.write(f"   {name}\n")

# === Main Script ===
def generate_docs_source():
    for base_module, output_dir in zip(BASE_MODULES, OUTPUT_DIRS):
        module_class_map = get_classes_grouped_by_module(base_module)
    
        for module_path, class_list in module_class_map.items():
            write_module_rst_file(module_path, class_list, output_dir)
    
        section_name = base_module.split('.')[-1].capitalize()
    
        write_index_rst_file(module_class_map, output_dir, section_name=section_name)
        print(f"✅ Documentation generated for {base_module} in {output_dir}")



#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import immutable_property, validate_args_type
from .s_ax import SAx
from typing import Self, Any, Callable
import numpy as np

class TAx(SAx):
	"""
	A Space to represent time axis.

	Note
	----
	- Use :class:`~modusa.generators.t_ax.TAxGen` API to instantiate this class.
	- It must be uniform with a well-defined sampling rate.
	- You are likely to be using this axis for most of the cases.
	- It is numpy compatible, so you can use numpy methods directly on this class object.
	- Since the object of this class represents time axis, any mathematical operations on it will result in another object of :class:`~modusa.models.tds.TDS` class with `y` being the result of the operation and `t` being the axis itself.
	
	Parameters
	----------
	n_points: int
		- Number of data points for the time axis.
	sr: float
		- Sampling rate.
		- Default: 1.0
	t0: float
		- Start timestamp.
		- Default: 0.0
	label: str
		- Label associated with the time axis.
		- Default: None => ''
		- e.g. (Time (sec))
	"""
	
	#--------Meta Information----------
	_name = "Time Axis"
	_nickname = "axis" # This is to be used in repr/str methods
	_description = "A Space to represent time axis."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-26"
	#----------------------------------
	
	@validate_args_type()
	def __init__(self, n_points, sr=1.0, t0=0.0, label=None):
		
		# Create `t` time series array
		t = t0 + np.arange(n_points) / sr
		
		super().__init__(values=t, label=label) # Instantiating `SAx` class
		
		# Storing other parameters so that we can use them.
		self._sr = sr
		self._t0 = t0
		

	#-----------------------------------
	# Properties (User Facing)
	#-----------------------------------
	
	@property
	def values(self) -> np.ndarray:
		return self._values
	
	@property
	def label(self) -> str:
		return self._label
	
	@property
	def shape(self) -> tuple:
		return self.values.shape
	
	@property
	def ndim(self) -> int:
		return self.values.ndim # Should be 1
	
	@property
	def size(self) -> int:
		return self.values.size
	
	@property
	def sr(self) -> float:
		return self._sr
	
	@property
	def t0(self) -> float:
		return self._t0
	
	@property
	def end_time(self) -> float:
		return float(self.values[-1])
	
	@property
	def duration(self) -> float:
		return float(self.end_time - self.t0)
	
	def __len__(self) -> int:
		return len(self.values)
	
	#===================================
	
	#------------------------------------
	# Utility methods
	#------------------------------------
	
	def is_same_as(self, other) -> bool:
		"""
		Compare it with another SAx object.

		Parameters
		----------
		other: SAx
			Another object to compare with.
		
		Returns
		-------
		bool
			True if same ow False

		Note
		----
		- We check the shape and all the values.
		- We are not checking the labels for now.
		"""
		
		if other.size == 1: # Meaning it is scalar
			return True
		
		axis1_arr = np.asarray(self)
		axis2_arr = np.asarray(other)
		
		if not isinstance(axis2_arr, type(axis1_arr)):
			return False
		if axis1_arr.shape != axis2_arr.shape:
			return False
		if not np.allclose(axis1_arr, axis2_arr):
			return False
		
		return True
	
	def copy(self) -> Self:
		"""
		Return a new copy of SAx object.
		
		Returns
		-------
		SAx
			A new copy of the SAx object.
		"""
		
		return self.__class__(n_points=self.shape[0], sr=self.sr, t0=self.t0, label=self.label)
	
	
	def set_meta_info(self, label):
		"""
		Set meta info for the axis.

		Parameters
		----------
		label: str
			Label for the axis (e.g. "Time (sec)").
		Returns
		-------
		Self
			A new Self instance with new label.
		
		.. code-block:: python
		
			import modusa as ms
			x = ms.sax.linear(100, 10)
			print(x)
			x = x.set_meta_info("My Axis (unit)")
			print(x)

			# I personally prefer setting it inline
			x = ms.sax.linear(100, 10).set_meta_info("My Axis (unit)")
			print(x)
		
		"""
		
		if label is None:
			return self
		else:
			return self.__class__(n_points=self.shape[0], sr=self.sr, t0=self.t0, label=label)
		
	def translate(self, n_samples):
		"""
		Translate the time axis by `n_samples`.

		Note
		----
		- `n_samples` can be both positive and negative.
		- You might end up getting -ve time values as we are not checking for the values rn.

		Parameters
		----------
		n_samples: int
			- Number of samples to move the signal.
			- +ve => moving signal forward.
			- -ve => moving signal backward.
		
		Returns
		-------
		TAx
			Translated axis.
		"""
		
		new_t0 = self.t0 + (n_samples / self.sr)
		
		return self.__class__(n_points=self.shape[0], sr=self.sr, t0=new_t0, label=self.label)
	
	#===================================
	
	
	#-------------------------------
	# NumPy Protocol
	#-------------------------------
	def __array__(self, dtype=None) -> np.ndarray:
		return np.asarray(self.values, dtype=dtype)
	
	def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
		"""
		Provides operation support for the universal functions
		on the TAx object.
		"""
		from .tds import TDS
		from .data import Data
		
		raw_inputs = [t.values if isinstance(t, type(self)) else t for t in inputs]
		
		# Call the actual ufunc
		result = getattr(ufunc, method)(*raw_inputs, **kwargs)
		
		if isinstance(result, (np.ndarray, np.generic)):
			y = Data(values=result, label=None)
			t = self
			return TDS(y=y, t=t, title=None)
		else:
			return result
		
	def __array_function__(self, func, types, args, kwargs):
		"""
		Additional numpy function support.
		"""
		from modusa.utils import np_func_cat as nfc
		from .data import Data
		
		if not all(issubclass(t, type(self)) for t in types):
			return NotImplemented
		
		# Not supporting concatenate like operations as axis any random axis can't be concatenated
		if func in nfc.CONCAT_FUNCS:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
			
		# Single signal input expected
		t = args[0]
		t_arr = np.asarray(t)
		result = func(t_arr, **kwargs)
		
		if func in nfc.REDUCTION_FUNCS:
			# If the number of dimensions is reduced
			if result.ndim == 0:
				return Data(values=result, label=None)
			else:
				raise RuntimeError(f"Unexpected result: `result` has more than 0 dimensions, {result.ndim}")
		
		elif func in nfc.X_NEEDS_ADJUSTMENT_FUNCS:
			# You must define logic for adjusting x
			raise NotImplementedError(f"{func.__name__} requires x-axis adjustment logic.")
			
		else:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")

	#================================
	
	
	#-----------------------------------
	# Indexing
	#-----------------------------------
		
	def __getitem__(self, key) -> Self:
		"""
		Defining how to index SAx instance.

		.. code-block:: python
			
			import modusa as ms
			x = ms.sax.linear(100, 10)
			print(x)
			print(x[10:20])

		Parameters
		----------
		key: int | slice
			What can go inside the square bracket [] for indexing.
		
		Returns
		-------
		SAx:
			Sliced instance of the axis.
		"""
		if not isinstance(key, (int, slice, tuple)):
			raise TypeError(f"Invalid key type {type(key)}")
		
		if isinstance(key, int):
			sliced_value = self.values[key]
			new_t0 = sliced_value
			return self.__class__(n_points=1, sr=self.sr, t0=new_t0, label=self.label)
			
		elif isinstance(key, slice):
			step = key.step or 1
			if step < 0:
				raise ValueError("Reversed slicing of time axis is not allowed.")
	
			sliced_values = self.values[key]
			new_n_points = len(sliced_values)
			new_sr = self.sr / step
			new_t0 = sliced_values[0]
		
			return self.__class__(n_points=new_n_points, sr=new_sr, t0=new_t0, label=self.label)
	
	
	def __setitem__(self, key, value):
		"""
		Raises error if trying to set values
		of an axis.

		Meaningful axis are not meant to be altered.
		"""
		raise TypeError("Time axis does not support item assignment.")
		
	#===============================


#!/usr/bin/env python3

from .audio_player import play
from .audio_converter import convert
from .youtube_downloader import download
from .audio_loader import load
from .audio_saver import save
from .ann_loader import load_ann
from .audio_recorder import record

from .plotter import Fig as fig
from .plotter import dist_plot, hill_plot, plot


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import validate_args_type
from modusa.generators.base import ModusaGenerator
from modusa.models.audio_signal import AudioSignal
import numpy as np

class AudioWaveformGenerator(ModusaGenerator):
	"""
	Generates different kinds of audio waveforms particulary helpful
	in teaching signal processing concepts and testing out newly
	created tools.
	"""
	
	#--------Meta Information----------
	_name = "Audio Waveform Generator"
	_description = "Generates different kind of audio waveforms."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-07"
	#----------------------------------
	
	@staticmethod
	def generate_example() -> "AudioSignal":
		"""
		Generates a simple sine wave audio signal as an example.
	
		Returns
		-------
		AudioSignal
			A 600 Hz sine wave lasting 2 seconds, sampled at 10,000 Hz.
		"""
		
		sr = 10000 # Hz
		duration = 2 # sec
		freq = 600 # Hz
		
		t = np.arange(0, duration, 1 / sr)
		y = np.sin(2 * np.pi * freq * t)
		
		signal = AudioSignal(y=y, sr=sr, title="Example")  # assuming AudioSignal accepts y and t
		
		return signal
	
	
	@staticmethod
	def generate_random(duration: float = 1.0, sr: int = 10000) -> "AudioSignal":
		"""
		Generates a random audio signal of given duration and sample rate.
	
		Parameters
		----------
		duration : float, optional
			Duration of the signal in seconds (default is 1.0).
		sr : int, optional
			Sampling rate in Hz (default is 10,000).
	
		Returns
		-------
		AudioSignal
			A randomly generated signal of the specified duration and sample rate.
		"""
		num_samples = int(duration * sr)
		t = np.linspace(0, duration, num=num_samples, endpoint=False)
		y = np.random.uniform(low=-1.0, high=1.0, size=num_samples)  # use uniform [-1, 1] for audio-like signal
		
		signal = AudioSignal(y=y, t=t, title="Random")
		
		return signal
	
	@staticmethod
	@validate_args_type()
	def generate_sinusoid(
		A: float | int = 1.0,
		f: float | int = 10.0,
		phi: float | int = 0.0,
		duration: float | int = 1.0,
		sr: int = 1000,
	) -> "AudioSignal":
		"""
		Generates a sinusoid audio signal with specified
		amplitude, frequency, phase, duration, and sample rate.
	
		Parameters
		----------
		A : float
			Amplitude of the sinusoid (default: 1.0)
		f : float
			Frequency in Hz (default: 10.0)
		phi : float
			Phase in radians (default: 0.0)
		duration : float
			Duration of the signal in seconds (default: 1.0)
		sr : int
			Sampling rate in Hz (default: 1000)
	
		Returns
		-------
		AudioSignal
			A sinusoidal signal with the given parameters.
		"""
		A, f, phi, duration, sr = float(A), float(f), float(phi), float(duration), int(sr)
		
		t = np.arange(0, duration, 1 / sr)
		y = A * np.sin(2 * np.pi * f * t + phi)
		
		signal = AudioSignal(y=y, sr=sr, title=f"Sinusoid ({f} Hz)")
		
		return signal
	
	@staticmethod
	@validate_args_type()
	def generate_square(
		A: float | int = 1.0,
		f: float | int = 10.0,
		phi: float | int = 0.0,
		duration: float | int = 1.0,
		sr: int = 1000,
	) -> "AudioSignal":
		"""
		Generates a square wave audio signal with specified
		amplitude, frequency, phase, duration, and sample rate.
	
		Parameters
		----------
		A : float
			Amplitude of the square wave (default: 1.0)
		f : float
			Frequency in Hz (default: 10.0)
		phi : float
			Phase in radians (default: 0.0)
		duration : float
			Duration of the signal in seconds (default: 1.0)
		sr : int
			Sampling rate in Hz (default: 1000)
	
		Returns
		-------
		AudioSignal
			A square wave signal of the specified parameters.
		"""
		A, f, phi, duration, sr = float(A), float(f), float(phi), float(duration), int(sr)
		t = np.arange(0, duration, 1 / sr)
		
		y = A * np.sign(np.sin(2 * np.pi * f * t + phi))
		
		signal = AudioSignal(y=y, sr=sr, title=f"Square ({f} Hz)")
		
		return signal
	
	
	@staticmethod
	@validate_args_type()
	def generate_sawtooth(
		A: float | int = 1.0,
		f: float | int = 10.0,
		phi: float | int = 0.0,
		duration: float | int = 1.0,
		sr: int = 1000,
	) -> "AudioSignal":
		"""
		Generates a sawtooth wave AudioSignal with specified amplitude, frequency, phase, duration, and sample rate.
	
		Parameters
		----------
		A : float
			Amplitude of the sawtooth wave (default: 1.0)
		f : float
			Frequency in Hz (default: 10.0)
		phi : float
			Phase in radians (default: 0.0)
		duration : float
			Duration of the signal in seconds (default: 1.0)
		sr : int
			Sampling rate in Hz (default: 1000)
	
		Returns
		-------
		AudioSignal
			A sawtooth wave signal of the specified parameters.
		"""
		A, f, phi, duration, sr = float(A), float(f), float(phi), float(duration), int(sr)
		t = np.arange(0, duration, 1 / sr)
		
		# Convert phase from radians to fractional cycle offset
		phase_offset = phi / (2 * np.pi)
		y = A * (2 * ((f * t + phase_offset) % 1) - 1)
		
		signal = AudioSignal(y=y, sr=sr, title=f"Sawtooth ({f} Hz)")
		return signal
	
	
	@staticmethod
	@validate_args_type()
	def generate_triangle(
		A: float | int = 1.0,
		f: float | int = 10.0,
		phi: float | int = 0.0,
		duration: float | int = 1.0,
		sr: int = 1000,
	) -> "AudioSignal":
		"""
		Generates a triangle wave AudioSignal with specified
		amplitude, frequency, phase, duration, and sample rate.
	
		Parameters
		----------
		A : float
			Amplitude of the triangle wave (default: 1.0)
		f : float
			Frequency in Hz (default: 10.0)
		phi : float
			Phase in radians (default: 0.0)
		duration : float
			Duration of the signal in seconds (default: 1.0)
		sr : int
			Sampling rate in Hz (default: 1000)
	
		Returns
		-------
		AudioSignal
			A triangle wave signal of the specified parameters.
		"""
		A, f, phi, duration, sr = float(A), float(f), float(phi), float(duration), int(sr)
		t = np.arange(0, duration, 1 / sr)
		phase_offset = phi / (2 * np.pi)  # Convert radians to cycle offset
		
		# Triangle wave formula: 2 * abs(2 * frac(x) - 1) - 1 scaled to amplitude
		y = A * (2 * np.abs(2 * ((f * t + phase_offset) % 1) - 1) - 1)
		
		signal = AudioSignal(y=y, sr=sr, title=f"Triangle ({f} Hz)")
		
		return signal
	


#!/usr/bin/env python3

from .base import ModusaSignalData, ModusaSignalAxis, ModusaSignal

from .data import Data
from .s_ax import SAx
from .t_ax import TAx

from .s1d import S1D
from .tds import TDS
from .audio import Audio

from .s2d import S2D
from .ftds import FTDS


#!/usr/bin/env python3

import inspect
from pathlib import Path

PLUGIN_PATH = Path(__file__).parent.parent / "plugins" # This directory contains all the plugins

def find_plugin_files():
	return [
			path for path in PLUGIN_PATH.rglob("*.py")
			if path.name not in {"__init__.py", "base.py"} # We do not want to show base plugins
		]

def load_plugin_class_from_file(file_path):
	import importlib.util
	from modusa.plugins.base import ModusaPlugin
	
	spec = importlib.util.spec_from_file_location(file_path.stem, file_path)
	module = importlib.util.module_from_spec(spec)
	try:
		spec.loader.exec_module(module)
	except Exception as e:
		print(f"❌ Error loading {file_path}: {e}")
		return []
	
	plugin_classes = []
	for _, obj in inspect.getmembers(module, inspect.isclass):
		if issubclass(obj, ModusaPlugin) and obj is not ModusaPlugin:
			plugin_classes.append(obj)
			
	return plugin_classes

def list_plugins():
	from rich.console import Console
	from rich.table import Table
	from modusa.plugins.base import ModusaPlugin
	
	console = Console()
	table = Table(title="🔌 Available Modusa Plugins")
	
	table.add_column("Plugin", style="bold green")
	table.add_column("Module", style="dim")
	table.add_column("Description", style="white")
	
	all_plugins = []
	
	for file_path in find_plugin_files():
		plugin_classes = load_plugin_class_from_file(file_path)
		for cls in plugin_classes:
			name = cls.__name__
			module = file_path.relative_to(PLUGIN_PATH.parent)
			author = getattr(cls, "author_name", "—")
			email = getattr(cls, "author_email", "—")
			desc = getattr(cls, "description", "—")
			table.add_row(name, str(module), desc)
			table.add_row("")
			all_plugins.append(cls)
			
	console.print(table)




#!/usr/bin/env python3

from IPython.display import Audio, HTML, display
from pathlib import Path
import numpy as np
import base64

def play(y, sr: float, clip=None, label=None):
	"""
	Audio player with optional clip selection, transcription-style label.

	Parameters
	----------
	y: ndarray
		- Audio signal.
	sr: float
		- Sampling rate.
	clip: tuple[float, float] | None
		- The portion from the audio signal to be played.
	label: str | None
		- Could be transcription/labels attached to the audio.
	
	Returns
	-------
	None
	"""
	start_time = 0.0
	end_time = len(y) / sr if y.ndim < 2 else y[0].size / sr
	
	# Optional clip selection
	if clip is not None:
		if not isinstance(clip, tuple) or len(clip) != 2:
			raise ValueError("`clip` must be a tuple of (start_time, end_time)")
		start_sample = int(clip[0] * sr)
		end_sample = int(clip[1] * sr)
		y = y[start_sample:end_sample]
		start_time, end_time = clip
		
	# Load and embed logo image as base64
	logo_path = Path(__file__).resolve().parent.parent / "images" / "icon.png"
	logo_size = 40
	logo_html = ""
	logo_link = "https://meluron.github.io/modusa"
	
	if logo_path.exists():
		with open(logo_path, "rb") as f:
			encoded_logo = base64.b64encode(f.read()).decode("utf-8")
			
		# Wrap logo in <a> so it's clickable
		logo_html = f"""
			<a href="{logo_link}" target="_blank" style="text-decoration:none;">
				<img src="data:image/png;base64,{encoded_logo}"
					style="
						position:absolute;
						bottom:8px;
						right:10px;
						width:{logo_size}px;
						height:{logo_size}px;
						opacity:0.8;
						transition:opacity 0.2s ease;
					"
					onmouseover="this.style.opacity=1.0"
					onmouseout="this.style.opacity=0.8"
				/>
			</a>
		"""
	audio_html = Audio(data=y, rate=sr)._repr_html_()
	
	label_html = f"""
		<div style="
			margin-top:4px;
			padding:10px 12px;
			background:#f7f7f7;
			border-radius:6px;
			color:#222;
			font-size:14px;
			line-height:1.5;
		">
			<strong>{start_time:.2f}s → {end_time:.2f}s:</strong>
			<span style="margin-left:6px;">{label if label else ''}</span>
		</div>
	"""
	
	html = f"""
	<div style="
		display:inline-block;
		position:relative;
		border:1px solid #e0e0e0;
		border-radius:10px;
		padding:14px 18px 36px 18px;
		background:#fff;
		font-family:sans-serif;
		max-width:520px;
		box-shadow:0 1px 3px rgba(0,0,0,0.05);
	">
		{label_html}
		<div style="margin-top:10px; margin-bottom:10px">
			{audio_html}
		</div>
		{logo_html}
	</div>
	"""
	
	display(HTML(html))
	


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import immutable_property, validate_args_type
from .s2d import S2D
from .s_ax import SAx
from .t_ax import TAx
from .data import Data
from modusa.tools.math_ops import MathOps
from typing import Self, Any, Callable
import numpy as np
import matplotlib.pyplot as plt
import modusa as ms

class FTDS(S2D):
	"""
	Space to represent feature time domain signal (2D).

	Note
	----
	- Use :class:`~modusa.generators.ftds.FTDSGen` API to instantiate this class.
	- The signal must have uniform time axis thus `TAx`.

	Parameters
	----------
	M: Data
		- Data object holding the main 2D array.
	f: SAx
		- Feature-axis of the 2D signal.
	t: TAx
		- Time-axis of the 2D signal.
	title: str
		- What does the signal represent?
		- e.g. "MySignal"
		- This is used as the title while plotting.
	"""
	
	#--------Meta Information----------
	_name = "Feature Time Domain Signal"
	_nickname = "FTDS" # This is to be used in repr/str methods
	_description = "Space to represent feature time domain signal (2D)."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-21"
	#----------------------------------
	
	def __init__(self, M, f, t, title = None):
		
		if not (isinstance(M, Data) and isinstance(f, SAx), isinstance(t, TAx)):
			raise TypeError(f"`M` must be `Data` instance, `f` and `x` must be `SAx` and `TAx` instances, got {type(M)}, {type(f)} and {type(t)}")
		
		super().__init__(M=M, y=f, x=t, title=title) # Instantiating `ModusaSignal` class
	
	#--------------------------------------
	# Properties
	#--------------------------------------
		
	@property
	def M(self) -> Data:
		return self._M
	
	@property
	def f(self) -> SAx:
		return self._y
	
	@property
	def t(self) -> SAx:
		return self._x
	
	@property
	def title(self) -> str:
		return self._title
	
	@property
	def shape(self) -> tuple:
		return self.M.shape
	
	@property
	def ndim(self) -> tuple:
		return self.M.ndim # Should be 2
	
	@property
	def size(self) -> int:
		return self.M.size
	
	#===================================
	
		
	#-------------------------------
	# NumPy Protocol
	#-------------------------------
	
	def __array__(self, dtype=None) -> np.ndarray:
		return np.asarray(self.M.values, dtype=dtype)
	
	def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
		"""
		Supports NumPy universal functions on the Signal1D object.
		"""
		from .data import Data  # Ensure this is the same Data class you're using
		from modusa.utils import np_func_cat as nfc
		
		raw_inputs = [
			np.asarray(obj.M) if isinstance(obj, type(self)) else obj
			for obj in inputs
		]
		
		result = getattr(ufunc, method)(*raw_inputs, **kwargs)
		
		result = Data(values=result, label=None)
		f = self.f.copy()
		t = self.t.copy()
		
		if result.shape[0] != f.shape[0] or result.shape[1] != t.shape[0]:
			raise ValueError(f"`{ufunc.__name__}` caused shape mismatch between data and axis, please create a github issue")
			
		return self.__class__(M=result, f=f, t=t, title=self.title)
	
	def __array_function__(self, func, types, args, kwargs):
		"""
		Additional numpy function support for modusa signals.
		Handles reduction and ufunc-like behavior.
		"""
		from .data import Data
		from modusa.utils import np_func_cat as nfc
		
		if not all(issubclass(t, type(self)) for t in types):
			return NotImplemented
		
		if func in nfc.CONCAT_FUNCS:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
			
		signal = args[0]
		result: Data = func(signal.M, **kwargs)
		axis = kwargs.get("axis", None)
		keepdims = kwargs.get("keepdims", None)
		
		if func in nfc.REDUCTION_FUNCS:
			if keepdims is None or keepdims is True:
				if axis is None: # Both axes collapsed
					dummy_f = SAx(0, "")
					dummy_t = TAx(n_points=1, sr=signal.t.sr, t0=0.0, label="")
					return self.__class__(M=result, f=dummy_f, t=dummy_t, title=signal.title)
			
				if isinstance(axis, int): # One of the axis collapsed
					if axis == 0:
						dummy_f = SAx(0, "")
						return self.__class__(M=result, f=dummy_f, t=signal.t.copy(), title=signal.title)
					elif axis in [1, -1]:
						dummy_t = TAx(n_points=1, sr=signal.t.sr, t0=0.0, label="")
						return self.__class__(M=result, f=signal.f.copy(), t=dummy_t, title=signal.title)
					else:
						raise ValueError
			elif keepdims is False:
				if axis is None: # Return Data
					return result
				if axis == 0: # Return TDS
					from .tds import TDS
					return TDS(y=result, t=signal.t, title=signal.title)
				elif axis in [1, -1]: # Return S1D
					from .s1d import S1D
					return S1D(y=result, x=signal.f, title=signal.title)
				else:
					raise ValueError
							
	
			# Case 3: Reduction keeps both axes (unlikely)
			else:
				raise NotImplementedError(f"{func.__name__} result shape={result.shape} not handled for modusa signal")
				
		elif func in nfc.X_NEEDS_ADJUSTMENT_FUNCS:
			raise NotImplementedError(f"{func.__name__} requires x-axis adjustment logic.")
			
		else:
			raise NotImplementedError(f"`{func.__name__}` is not yet tested on modusa signal, please create a GitHub issue.")
			
			
	#================================
	
	#-------------------------------
	# Indexing
	#-------------------------------
			
	def __getitem__(self, key):
		"""
		Return a sliced or indexed view of the data.
		
		Parameters
		----------
		key : int | slice | S2D
			- Index to apply to the values.
		
		Returns
		-------
		S2D | S1D | Data
			- S2D object if slicing results in 2D array
			- S1D object if slicing results in 1D array
			- Data if slicing results in scalar
		"""
		from .s1d import S1D
		from .tds import TDS
		
		if isinstance(key, S1D):
			raise TypeError(f"Applying `S1D` mask on `S2D` is not allowed.")
			
		# We slice the data
		sliced_M = self.M[key]
		
		# Case 1: Row indexing only — return a horizontal slice (1D view across columns)
		if isinstance(key, int):
			sliced_f = self.f[key]
			sliced_t = self.t
			return TDS(y=sliced_M, x=sliced_t, title=self.title)
		
		# Case 2: Column indexing only — return a vertical slice (1D view across rows)
		elif isinstance(key, slice):
			sliced_f = self.f[key]
			sliced_t = self.t
			return self.__class__(M=sliced_M, f=sliced_f, t=sliced_t, title=self.title)
		
		# Case 3: 2D slicing
		elif isinstance(key, tuple) and len(key) == 2:
			row_key, col_key = key
			if isinstance(row_key, int) and isinstance(col_key, int):
				# Single value extraction → shape = (1, 1)
				# Will return data object
				return Data(values=sliced_M, title=self.title)
			
			elif isinstance(row_key, int) and isinstance(col_key, slice):
				# Row vector → return S1D
				sliced_f = self.f[row_key] # Scalar
				sliced_t = self.t[col_key]
				
				return TDS(y=sliced_M, t=sliced_t, title=self.title)
			
			elif isinstance(row_key, slice) and isinstance(col_key, int):
				# Column vector → return S1D
				sliced_f = self.f[row_key]
				sliced_t = self.t[col_key] # Scalar
				
				return S1D(y=sliced_M, x=sliced_f, title=self.title)
			
			elif isinstance(row_key, slice) and isinstance(col_key, slice):
				# 2D slice → return same class
				sliced_f = self.f[row_key]
				sliced_t = self.t[col_key] 
				
				return self.__class__(M=sliced_M, f=sliced_f, t=sliced_t, title=self.title)
		
		# Case 4: Boolean masking signal
		elif isinstance(key, type(self)):
			sliced_f = self.f
			sliced_t = self.t
			
			return self.__class__(M=sliced_M, f=sliced_f, t=sliced_t, title=self.title)
		
		else:
			raise TypeError(f"Unsupported index type: {type(key)}")
			
	def __setitem__(self, key, value):
		"""
		Set values at the specified index.
	
		Parameters
		----------
		key : int | slice | array-like | boolean array | S1D
			Index to apply to the values.
		value : int | float | array-like
			Value(s) to set.
		"""
		
		self.M[key] = value  # In-place assignment
		
	#===================================
		
	#-------------------------------
	# Basic arithmetic operations
	#-------------------------------
	def __add__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.add(self, other) 
	
	def __radd__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.add(other, self)
	
	def __sub__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.subtract(self, other)
	
	def __rsub__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.subtract(other, self)
	
	def __mul__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.multiply(self, other) 
	
	def __rmul__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.multiply(other, self)
	
	def __truediv__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.divide(self, other) 
	
	def __rtruediv__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.divide(other, self)
	
	def __floordiv__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.floor_divide(self, other) 
	
	def __rfloordiv__(self, other):
		if not self.has_same_axis_as(other):
			raise ValueError("Axes are not aligned for the operation.")
		return np.floor_divide(other, self)
	
	def __pow__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.power(self, other) 
	
	def __rpow__(self, other):
		if isinstance(other, type(self)):
			if not self.has_same_axis_as(other):
				raise ValueError("Axes are not aligned for the operation.")
		return np.power(other, self)
	
	#===============================
	
	
	#-------------------------------
	# Basic comparison operations
	#-------------------------------
	def __eq__(self, other):
		return np.equal(self, other)
	
	def __ne__(self, other):
		return np.not_equal(self, other)
	
	def __lt__(self, other):
		return np.less(self, other)
	
	def __le__(self, other):
		return np.less_equal(self, other)
	
	def __gt__(self, other):
		return np.greater(self, other)
	
	def __ge__(self, other):
		return np.greater_equal(self, other)
	
	#====================================
	
	
	#-----------------------------------
	# Utility Methods
	#-----------------------------------
	
	def unpack(self):
		"""
		Unpacks the object into easy to work
		with data structures.

		Returns
		-------
		(np.ndarray, np.ndarray, np.ndarray)
			- M: Signal data array.
			- f: Signal feature-axis array.
			- t: Signal time-axis array.
		"""
		
		M = self.M.values
		f = self.f.values
		t = self.t.values
		
		return (M, f, t)
	
	def copy(self) -> Self:
		"""
		Returns a new copy of the signal.

		Returns
		-------
		Self
			A new copy of the object.
		"""
		
		copied_M = self.M.copy()
		copied_f = self.f.copy()
		copied_t = self.t.copy()
		title = self.title # Immutable, hence no need to copy
		
		return self.__class__(M=copied_M, f=copied_f, t=copied_t, title=title)
	
	def set_meta_info(self, title = None, M_label = None, f_label = None, t_label = None) -> None:
		"""
		Set meta info about the signals.

		Parameters
		----------
		title: str
			- Title for the signal
			- e.g. "MyTitle"
		M_label: str
			- Label for the data that matrix is holding.
			- e.g. "Intensity (dB)"
		f_label: str
			- Label for the feature-axis.
			- e.g. "Frequency (Hz)"
		t_label: str
			- Label for the time-axis.
			- e.g. "Time (sec)"
		Returns
		-------
		FTDS
			A new instance with updated meta info.
		"""
		
		M, f, t = self.M, self.f, self.t
		
		M_label = str(M_label) if M_label is not None else M.label
		f_label = str(f_label) if f_label is not None else f.label
		t_label = str(t_label) if t_label is not None else t.label
		title = str(title) if title is not None else self.title
		
		# We create a new copy of the data and axis
		new_M = M.set_meta_info(label=M_label)
		new_f = f.set_meta_info(label=f_label)
		new_t = t.set_meta_info(label=t_label)
		
		return self.__class__(M=new_M, f=new_f, t=new_t, title=title)
	
	def translate_t(self, n_samples):
		"""
		Translate the FTDS signal along time axis
		by `n_samples`.

		Note
		----
		- `n_samples` can be both positive and negative.
		- You might end up getting -ve time values as we are not checking for the values rn.

		Parameters
		----------
		n_samples: int
			- Number of samples to move the signal.
			- +ve => moving signal forward.
			- -ve => moving signal backward.
		
		Returns
		-------
		FTDS
			Translated signal.
		"""
		
		# We just need to create a new time axis with shifted t0
		translated_t = self.t.translate(n_samples=n_samples)
		
		translated_signal = self.__class__(M=self.M.copy(), f=self.f.copy(), t=translated_t, title=self.title)
		
		return translated_signal
	
	
	def mask(self, condition, set_to=None) -> Self:
		"""
		Mask the signal based on condition and
		the values can be set.
		
		Parameters
		----------
		condition: Callable
			- Condition function to apply on values of the signal.
			- E.g. lambda x: x > 10
		set_to: Number
			- Number to replace the masked position values.

		Returns
		-------
		S2D
			Masked Signal
		"""
		
		mask = condition(self)
		new_val = set_to
		
		if set_to is None: # Return the mask as the same signal but with booleans
			return mask
	
		else:
			# We apply the mask and update the signal data
			new_data = self.M.mask(condition=condition, set_to=new_val)
			
			# Since we're just updating the data, there is no change in the axis
			return self.__class__(M=new_data, f=self.f.copy(), t=self.t.copy(), title=self.title)
		
		
	def pad(self, left=None, right=None) -> Self:
		"""
		Pad the signal with array like object from the
		left or right.

		Parameters
		----------
		left: arraylike
			- What to pad to the left of the signal.
			- E.g. 1 or [1, 0, 1], np.array([1, 2, 3])
		right: arraylike
			- What to pad to the right of the signal.
			- E.g. 1 or [1, 0, 1], np.array([1, 2, 3])

		Returns
		-------
		FTDS
			Padded signal.
		"""
		
		if right is None and left is None: # No padding applied
			return self
	
		# Pad the data
		M_padded = self.M.pad(left=left, right=right)
	
		# Find the new t0
		if left is not None:
			if np.ndim(left) == 0: left = np.asarray([left])
			else: left = np.asarray(left)
			new_t0 = self.t.t0 - (left.shape[0] / self.t.sr)
		else:
			new_t0 = self.t.t0

		t_padded = self.t.__class__(n_points=M_padded.shape[1], sr=self.t.sr, t0=new_t0, label=self.t.label)
		
		return self.__class__(M=M_padded, f=self.f.copy(), t=t_padded, title=self.title)
	
	#====================================
	
	
	
	
	#-----------------------------------
	# Info
	#-----------------------------------
	
	def print_info(self) -> None:
		"""Print key information about the FTDS signal."""
		print("-"*50)
		print(f"{'Title':<20}: {self.title}")
		print("-"*50)
		print(f"{'Type':<20}: {self.__class__.__name__}")
		print(f"{'Shape':<20}: {self.shape} (freq bins × time frames)")
		print(f"{'Duration':<20}: {self.t.duration}")
		print(f"{'Frame Rate':<20}: {self.t.sr} (frames / sec)")
		print(f"{'Frame Duration':<20}: {1 / self.t.sr:.4f} sec ({(1 / self.t.sr) * 1000:.2f} ms)")
		
		# Inheritance chain
		cls_chain = " → ".join(cls.__name__ for cls in reversed(self.__class__.__mro__[:-1]))
		print(f"{'Inheritance':<20}: {cls_chain}")
		print("=" * 50)
		
		
	#===================================


#!/usr/bin/env python3


from abc import ABC, abstractmethod

class ModusaSignal(ABC):
	"""
	Base class for any signal in the modusa framework.
	
	Note
	----
	- Intended to be subclassed.
	"""
	
	#--------Meta Information----------
	_name = "Modusa Signal"
	_nickname = "MS" # This is to be used in repr/str methods
	_description = "Base class for any signal types in the Modusa framework."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-06-23"
	#----------------------------------
	
	def __init__(self):
		pass
		

class ModusaSignalAxis(ABC):
	"""
	Base class for any modusa signal axis in the modusa framework.
	
	Note
	----
	- Intended to be subclassed.
	"""
	
	#--------Meta Information----------
	_name = "Modusa Signal Axis"
	_nickname = "MSAx" # This is to be used in repr/str methods
	_description = "Base class for any modusa signal axis in the modusa framework."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-25"
	#----------------------------------
	
	def __init__(self):
		pass
	
class ModusaSignalData(ABC):
	"""
	Base class for any modusa signal data in the modusa framework.
	
	Note
	----
	- Intended to be subclassed.
	"""
	
	#--------Meta Information----------
	_name = "Modusa Signal Data"
	_nickname = "MSData" # This is to be used in repr/str methods
	_description = "Base class for any modusa signal data in the modusa framework."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-27"
	#----------------------------------
	
	def __init__(self):
		pass

	


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import validate_args_type
from .base import ModusaGenerator
from modusa.models.s_ax import SAx
import numpy as np

class SAxGen(ModusaGenerator):
	"""
	Provides user friendly APIs to generate axis for
	signals (instances of `SAx`).
	"""
	
	#--------Meta Information----------
	_name = "SignalAxisGenerator"
	_description = "APIs to generate axis for signals."
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-25"
	#----------------------------------
	
	
	@classmethod
	@validate_args_type()
	def from_array(
		cls,
		values: np.ndarray | list | float | int | np.generic,
		label: str = "SAx"
	) -> SAx:
		"""
		Create `SAx` instance from basic data structures.
		
		.. code-block:: python

			import modusa as ms
			x = ms.sax.from_array([1, 2, 3])
			print(x)
			time_sax.print_info()

		Parameters
		----------
		values: np.ndarray | list | float | int | np.generic
			- The values for the axis.
		label: str
			- Label for the axis.

		Returns
		-------
		SAx
			An instance of SAx.
		"""
		
		if isinstance(values, (int, float, np.generic)): values = [values] # Scalar to 1D
		values = np.asarray(values)

		return SAx(values=values, label=label)
		
	
	@classmethod
	def linear(cls, n_points: int, sr: int | float = 1.0, start: int | float = 0.0, label: str = "Linear Axis") -> SAx:
		"""
		Create a linearly spaced axis.
		
		.. code-block:: python

			import modusa as ms
			x = ms.sax.linear(n_points=100, sr=2, start=10, label="Time (sec)")
			print(x)
			x.print_info()
		
		Parameters
		----------
		n_points: int
			- Number of data points for the axis.
		sr: int | float
			- Sampling rate of the axis.
		start: int | float
			- Start value.
		label: str
			- Label for the axis.
		
		Returns
		-------
		SAx
			An instance of SAx.
		"""
		
		assert isinstance(n_points, int)
		assert isinstance(sr, (int, float))
		assert isinstance(start, (int, float))
		assert isinstance(label, str)
	
		sr = float(sr)
		start = float(start)
	
		values = start + np.arange(n_points) / sr  # ensures exact number of points
		time_ax = SAx(values=values, label=label)
		time_ax.sr = sr
	
		return time_ax


#!/usr/bin/env python3


from modusa import excp
from modusa.decorators import validate_args_type
from .base import ModusaGenerator
from modusa.models.data import Data
from modusa.models.s_ax import SAx
from modusa.models.t_ax import TAx
from modusa.models.ftds import FTDS
import numpy as np

class FTDSGen(ModusaGenerator):
	"""
	Provides user friendly APIs to generate instances of different 
	`FTDS` instances.
	"""
	
	#--------Meta Information----------
	_name = ""
	_description = ""
	_author_name = "Ankit Anand"
	_author_email = "ankit0.anand0@gmail.com"
	_created_at = "2025-07-27"
	#----------------------------------
	
	@staticmethod
	def from_array(
		M: np.ndarray | list | float | int | np.generic,
		f: np.ndarray | list | float | int | np.generic | None = None,
		sr: int | float = 1.0,
		t0: int | float = 0.0,
		M_label: str = "M",
		f_label: str = "Feature",
		t_label: str = "Time (sec)",
		title: str = "Feature Time Domain Signal"
	) -> FTDS:
		"""
		Create `FDTS` instance from basic data structures.

		.. code-block:: python

			import modusa as ms
			M = ms.ftds.from_array([1, 2, 3])
			print(M)
			M.print_info()
		
		Parameters
		----------
		M: np.ndarray | list | float | int | np.generic
			- Data values.
		f: np.ndarray | list | float | int | np.generic | None
			- y axis values.
			- Default: None → Creates an integer indexing.
		sr: int | float
			- Sampling rate / Frame rate.
			- Default: 1.0
		t0: int | float
			- Start timestamp.
			- Default: 0.0
		M_label: str
			- Label for the data.
			- Default: "M"
		f_label: str
			- Feature label for the signal.
			- Default: "Feature"
		t_label: str
			- Time label for the signal.
			- Default: "Time (sec)"
		title: str
			- Title for the signal.
			- Default: "Feature Time Domain Signal"
		Returns
		-------
		FTDS
			An instance of FTDS.
		"""
		assert isinstance(M, (np.ndarray, list, float, int, np.generic))
		assert isinstance(f, (np.ndarray, list, float, int, np.generic)) or f is None
		assert isinstance(sr, (int, float)) and isinstance(t0, (int, float))
		assert isinstance(M_label, str) and isinstance(f_label, str) and isinstance(t_label, str) and isinstance(title, str)
		
		if isinstance(M, (float, int, np.generic)): M = [[M]] # Convert to list of 1 element
		if isinstance(f, (float, int, np.generic)): f = [f] # Convert to list of 1 element
		
		M = np.asarray(M)
		assert M.ndim == 2
		
		if f is None: f = np.arange(M.shape[0])
		else: f = np.asarray(f)
		assert f.ndim == 1
		assert f.shape[0] == M.shape[0], "Shape mismatch"
		
		sr = float(sr)
		t0 = float(t0)
		
		M = Data(values=M, label=M_label)
		f = SAx(values=f, label=f_label)
		t = TAx(n_points=M.shape[1], sr=sr, t0=t0, label=t_label)
		
		return FTDS(M=M, f=f, t=t, title=title)
	
	@classmethod
	def zeros(cls, shape, f=None, sr=1.0, t0=0.0) -> FTDS:
		"""
		Create `FTDS` instance with all zeros.

		.. code-block:: python

			import modusa as ms
			M = ms.ftds.zeros((10, 5))
			print(M)
			M.print_info()
		
		Parameters
		----------
		shape: tuple[int, int]
			- Shape of the signal with zeros.
			- Must be 1 dimensional
			- E.g. (10, 5)
		Returns
		-------
		FTDS
			An instance of FTDS.
		"""
		assert isinstance(shape, tuple)
		M = np.zeros(shape)
		
		return cls.from_array(M=M, f=f, sr=sr, t0=t0, title="Zeros")
	
	@classmethod
	def zeros_like(cls, signal: FTDS) -> FTDS:
		"""
		Create `FTDS` instance similar to `signal`
		but with all entries being zeros.

		.. code-block:: python

			import modusa as ms
			signal = ms.ftds.from_array([[1, 2, 3], [4, 5, 6]])
			M = ms.ftds.zeros_like(signal)
			print(M)
			M.print_info()
		
		Parameters
		----------
		signal: FTDS
			- Reference signal to create zeros like that.
		Returns
		-------
		FTDS
			An instance of FTDS.
		"""
		
		assert signal.__class__ in [FTDS]
		
		M = np.zeros(signal.shape)
		f = signal._f
		t = signal._t
		
		M_label = signal._M_label
		f_label = signal._f_label
		t_label = signal._t_label
		title = signal._title
		
		return cls.from_array(M=M, f=f, t=t, M_label=M_label, f_label=f_label, t_label=t_label, title=title)
	
	
	@classmethod
	def ones(cls, shape: tuple[int, int]) -> FTDS:
		"""
		Create `FTDS` instance with all ones.

		.. code-block:: python

			import modusa as ms
			M = ms.ftds.ones((10, 5))
			print(M)
			M.print_info()
		
		Parameters
		----------
		shape: tuple[int, int]
			- Shape of the signal with ones.
			- Must be 1 dimensional
			- E.g. (10, 5)
		Returns
		-------
		FTDS
			An instance of FTDS.
		"""
		assert isinstance(shape, tuple)
		M = np.ones(shape)
		
		return cls.from_array(M=M, title="Ones")
	
	@classmethod
	def ones_like(cls, signal: FTDS) -> FTDS:
		"""
		Create `FTDS` instance similar to `signal`
		but with all entries being ones.

		.. code-block:: python

			import modusa as ms
			signal = ms.ftds.from_array([[1, 2, 3], [4, 5, 6]])
			M = ms.ftds.ones_like(signal)
			print(M)
			M.print_info()
		
		Parameters
		----------
		signal: FTDS
			- Reference signal to create ones like that.
		Returns
		-------
		FTDS
			An instance of FTDS.
		"""
		
		assert signal.__class__ in [FTDS]
		
		M = np.ones(signal.shape)
		f = signal._f
		t = signal._t
		
		M_label = signal._M_label
		f_label = signal._f_label
		t_label = signal._t_label
		title = signal._title
		
		return cls.from_array(M=M, f=f, t=t, M_label=M_label, f_label=f_label, t_label=t_label, title=title)
	
	@classmethod
	def random(cls, shape: tuple[int, int]) -> FTDS:
		"""
		Create `FTDS` instance with random entries.

		.. code-block:: python

			import modusa as ms
			M = ms.ftds.random((10, 5))
			print(M)
			M.print_info()
		
		Parameters
		----------
		shape: tuple[int, int]
			- Shape of the signal with random values.
			- Must be 1 dimensional
			- E.g. (10, 5)
		Returns
		-------
		FTDS
			An instance of FTDS.
		"""
		assert isinstance(shape, tuple)
		M = np.random.random(shape)
		
		return cls.from_array(M=M, title="Random")
	
	@classmethod
	def random_like(cls, signal: FTDS) -> FTDS:
		"""
		Create `FTDS` instance similar to `signal`
		but with random entries.

		.. code-block:: python

			import modusa as ms
			signal = ms.ftds.from_array([[1, 2, 3], [4, 5, 6]])
			M = ms.ftds.random_like(signal)
			print(M)
			M.print_info()
		
		Parameters
		----------
		signal: FTDS
			- Reference signal.
		Returns
		-------
		FTDS
			An instance of FTDS with random values.
		"""
		
		assert signal.__class__ in [FTDS]
		
		M = np.random.random(signal.shape)
		f = signal._f
		t = signal._t
		
		M_label = signal._M_label
		f_label = signal._f_label
		t_label = signal._t_label
		title = signal._title
		
		return cls.from_array(M=M, f=f, t=t, M_label=M_label, f_label=f_label, t_label=t_label, title=title)
	